import unittest
import pytest
from pysercomb.parsers import racket
from pysercomb.parsing import float_, int_, scientific_notation
from pysercomb.pyr import units as pyru


class TestForms(unittest.TestCase):
    def test_int(self):
        res = int_('1')
        _, v, _  = res
        assert v == 1, res

    def test_int_leading_zero(self):
        res = int_('00001')
        _, v, _  = res
        assert v == 1, res

        res = int_('-00001')
        _, v, _  = res
        assert v == -1, res

    def test_scinot(self):
        _, v, _ = scientific_notation('1E10')
        assert v == 1E10, v
        assert repr(v) == '1E10', res

    def test_scinot_leading_zero(self):
        # REMINDER: the primary use case for this format
        # is as a means to record the exact notation used
        # by an investigator, therefore we can't parse this
        # to a float, HOWEVER this induces a roundtripping
        # issue when a float that is _not_ provided in
        # scientific notation is converted to scientific notation
        res = _, v, _ = scientific_notation('2e-06')
        assert v == 2e-06, res
        assert repr(v) == '2e-06', res

    def test_scinot_neg(self):
        res = scientific_notation('1')
        _, v, rest = res
        assert v is None and rest == '1', (v, rest)

    def test_quote(self):
        res = racket.racket_doc("'(a (1 . 2))")
        print(res)
        _, (v,), _ = res
        assert v == ('quote', ('a', (1, 2)))

    def test_cons_literal(self):
        _, (v,), _ = racket.racket_doc("'(1 . 2)")
        assert v == ('quote', (1 , 2))

    def test_null_t_f(self):
        _, (v,), _ = racket.racket_doc("'(null #t #f)")
        assert v == ('quote', ('null', True, False))

    def test_empty_list(self):
        _, (v,), _ = racket.racket_doc("'()")
        assert v is None

    def test_number(self):
        res = racket.racket_doc('1')
        _, (v,), _ = res 
        assert v == 1, res

    def test_null(self):
        code = """#lang racket/base
(provide something)
(define something
'((a . b)
  (() . 0)
  (c . 1)))
"""
        res = racket.racket_doc(code)
        print(res)
        _, v, _ = res
        assert v == ('module-unexp', None, ('racket/base',),
                     ('module-begin',
                      ('provide', 'something'),
                      ('define', 'something',
                       ('quote', (('a', 'b'),
                                  (tuple(), 0),
                                  ('c', 1))))))

    def test_string_escape(self):
        ok, value, rest = racket._string_escape(r'\"')
        assert ok

    def test_string_escape_full(self):
        code = r'''"testing \"string\" escape"'''
        res = racket.exp(code)
        print(res)
        _, v, _ = res
        assert v == 'testing "string" escape'

    def test_quote_sexp(self):
        code = """'(TEAPOTS ON THE OH NO)"""
        res = racket.exp(code)
        print(res)
        _, v, _ = res
        assert v == ('quote', ('TEAPOTS', 'ON', 'THE', 'OH', 'NO'))

    def test_quote_has_number(self):
        code = """'a4sdf"""
        res = racket.exp(code)
        print(res)
        _, v, _ = res
        assert v == ('quote', 'a4sdf')

    def test_quote_start_number(self):
        code = """'4asdf"""
        res = racket.exp(code)
        print(res)
        _, v, _ = res
        assert v == ('quote', '4asdf')

    def test_number_literal_fail(self):
        code = """4asdf"""
        ok, v, rest = res = racket.literal(code)
        print(res)
        ok, v, _ = res
        assert not ok

    def test_number_literal_ends(self):
        codes = [
            """4(+ 1 2)""",
            """4;lol""",
            """4)""",
            """4'hello""",
            """4""",
            '''4"there"'''
        ]
        for code in codes:
            ok, v, rest = res = racket.literal(code)
            print(res)
            assert ok and v == 4 and (code == '4' or rest)

    def test_multi_comment_exp(self):
        code = '''; c1
; c2
; how many HAHA! third times the charm
(+ 1  ; asdf
   #:k-w 3 ;oh no
   2) ; asdf
; c3
; c4'''
        ok, v, rest = res = racket.exp(code)
        assert ok and v == ('+', 1, '#:k-w', 3, 2), res

    @pytest.mark.skip('not implemented yet')
    def test_here_string(self):
        code = '''#<<abcdefg
yay some text!
abcdefg
'''
        racket.exp(code)
        ok, v, rest = res = racket.exp(code)
        assert ok and not rest, res


class TestPyrRacket(unittest.TestCase):

    def test_quote_sexp(self):
        hrm = pyru.RacketParser("'(i am a teapot short and stout)")
        ap = hrm.asPython()
        assert ap == ('quote', ('i', 'am', 'a', 'teapot', 'short', 'and', 'stout'))

    def test_quote_str(self):
        hrm = pyru.RacketParser(""" '"i am redundant" """)
        ap = hrm.asPython()
        assert ap == "i am redundant"

    def test_add(self):
        hrm = pyru.RacketParser('(+ 1 2)')
        ap = hrm.asPython()
        assert ap == 3

    def test_complex(self):
        test = """
(protc:executor-verb "diluted" (hyp: 'ycN7aDCqEemsREea4BwMxg)  ; https://hyp.is/ycN7aDCqEemsREea4BwMxg
  (protc:input "blocking buffer" (hyp: 'L9TvjkRTEem4q5dY1oVi-w))  ; https://hyp.is/L9TvjkRTEem4q5dY1oVi-w
  (protc:input "goat anti-VR1" (hyp: 'D91m9ippEemlRi84HUXyKw)  ; https://hyp.is/D91m9ippEemlRi84HUXyKw
    (protc:implied-aspect "dilution" (hyp: 'uN3DpiP6Eemt8lPtBr6wUg)  ; https://hyp.is/7nDe9iDOEem6X6NI4c_F3A
      (protc:invariant (param:ratio 1 150) (hyp: '7nDe9iDOEem6X6NI4c_F3A))))  ; https://hyp.is/7nDe9iDOEem6X6NI4c_F3A
  (protc:invariant (protc:fuzzy-quantity "overnight" "duration")
                    (hyp: 'B9_WniDPEemNW__FACccsA))  ; https://hyp.is/B9_WniDPEemNW__FACccsA
  (protc:parameter* (param:quantity 4 (param:unit 'degrees-celsius)) (hyp: 'DOeIYiDPEem7zPuTeuEhyA)))  ; https://hyp.is/DOeIYiDPEem7zPuTeuEhyA"""
        hrm = pyru.RacketParser(test)
        ap = hrm.asPython()

    def test_here(self):
        # TODO check behvior wrt EOF vs \nEOF
        test = """#<<----
    *--------------- ...
    |protc: https://hyp.is/--hiOoDkEeqlArtUO_9vSQ protc.byId('--hiOoDkEeqlArtUO_9vSQ')
    |value: ganglion neurons
    |exact: ganglion neurons
    |tags:  {'UBERON:0000045', 'sparc:AnatomicalLocation', 'protc:black-box'}
    |___________________________________________________________________________
----
"""
        hrm = pyru.RacketParser(test)
        ap = hrm.asPython()
        assert ap == test[8:-6]

    def test_number(self):
        test = '1'
        hrm = pyru.RacketParser(test)
        ap = hrm.asPython()
        assert ap == 1

        test = '1.0'
        hrm = pyru.RacketParser(test)
        ap = hrm.asPython()
        assert ap == 1.0

        test = '1e-10'
        hrm = pyru.RacketParser(test)
        ap = hrm.asPython()
        assert ap == 1e-10

    def test_liter_lambda_bug(self):
        test = """
(protc:parameter* (param:quantity 1 (param:unit 'liters 'micro)) #:prov (hyp: 'EQrx4CDKEemxBCfTHP1Bhg))  ; https://hyp.is/EQrx4CDKEemxBCfTHP1Bhg"""
        hrm = pyru.RacketParser(test)
        ap = hrm.asPython()
        assert ap.quantity.units.json() != 'lambda'



#!/usr/bin/env python3

import parsec
import os
from pysercomb.utils import coln
from IPython import embed

def param(symbol):
    def decorator(parser):
        return parser.parsecmap(lambda v: ('param:' + symbol, *v))
    return decorator

def BOX(v):
    return v,

def OR(parser, *parsers):
    if not parsers:
        return parser
    return parsec.try_choice(parser, OR(*parsers))

DEGREES_UNDERLINE = b'\xc2\xba'.decode()  # º sometimes pdfs misencode these
DEGREES_FEAR = b'\xe2\x97\xa6' # this thing is scary and I have no id what it is or why it wont change color ◦
def get_quoted_list(filename):
    with open(os.path.expanduser('~/git/protc/protc-lib/protc/units/' + filename), 'rt') as f:
        lines = [_.split(';')[0].strip() for _ in f.readlines()][3:-1]
    return [line.strip("'").strip('(').rstrip(')').split(' . ') for line in lines if line and '.' in line]

_SIPREFS, _SIEXPS, _SIUNITS, _EXTRAS = [get_quoted_list(_) for _ in ('si-prefixes-data.rkt', 'si-prefixes-exp-data.rkt', 'si-units-data.rkt', 'si-units-extras.rkt')]
_silookup = {k: "'" + v for k, v in
             _SIUNITS + _EXTRAS +
             [[v, v] for k, v in _SIUNITS] +
             [[v, v] for k, v in _EXTRAS]}
_siplookup = {k: "'" + v for k, v in _SIPREFS}

def make_funcs(inpt, lookuptable):
    args = []
    for token in sorted(inpt, key=lambda a: -len(a)):  # right associativity (ie da recognized even if d a token)
        args.append(parsec.string(token).parsecmap(lambda v: lookuptable[v]))
    return args

space = parsec.space()
spaces = parsec.spaces()
spaces1 = parsec.many1(space)
carrot = parsec.string('^')
exponent = carrot.parsecmap(lambda v: 'exponent')
division = parsec.string('/')
multiplication = parsec.string('*')
unit_op = division ^ multiplication
colon = parsec.string(':')
plus_or_minus_symbol = parsec.string('±')
plus_or_minus_pair = parsec.string('+-')  # yes that is an b'\x2d'
plus_or_minus = OR(plus_or_minus_symbol, plus_or_minus_pair).parsecmap(lambda v: 'plus-or-minus')
# I hate the people who felt the need to make different type blocks for this stuff in 1673
EN_DASH = b'\xe2\x80\x93'.decode()
HYPHEN_MINUS = b'\x2d'.decode()  # yes, the thing that most keyboards have
en_dash = parsec.string(EN_DASH)
hyphen_minus = parsec.string(HYPHEN_MINUS)
dash_thing = (en_dash ^ hyphen_minus).parsecmap(lambda v: HYPHEN_MINUS)  # THERE ARE TOO MANY AND THEY ALL LOOK THE SAME
double_dash_thing = parsec.times(dash_thing, 2)
thing_accepted_as_a_dash = OR(double_dash_thing, dash_thing).parsecmap(lambda v: HYPHEN_MINUS)
lt = parsec.string('<')
lte = parsec.string('<=')
gt = parsec.string('>')
gte = parsec.string('>=')
comparison = OR(lte, gte, lt, gt)
CROSS = b'\xc3\x97'.decode()
cross = parsec.string(CROSS)
x = parsec.string('x')
by = OR(cross, x)
approx = parsec.string('~').parsecmap(lambda v: 'approximately')
# number words
_numlookup = {
    'zero':0,
    'one':1,
    'two':2,
    'three':3,
    'four':4,
    'five':5,
    'six':6,
    'seven':7,
    'eight':8,
    'nine':9, }
num_word_lower = OR(*make_funcs(_numlookup, _numlookup))
@parsec.Parser
def num_word_cap(text, index):
    return num_word_lower(text.lower(), index)
num_word = num_word_lower ^ num_word_cap

# note that the behavior of these differes from mine in that they do not fail if we exceed max
def AT_MOST_ONE(parser): return parsec.times(parser, 0, 1).parsecmap(lambda v: v[0] if v else '')
def EXACTLY_ONE(parser): return parsec.times(parser, 1, 1).parsecmap(lambda v: v[0])

def join(v): return ''.join(v)

digit = parsec.digit()
digits = parsec.many(digit).parsecmap(lambda v: ''.join(v))
digits1 = parsec.many1(digit).parsecmap(lambda v: ''.join(v))
point = parsec.string('.')
_int_ = (AT_MOST_ONE(dash_thing) + digits1).parsecmap(join)
int_ = _int_.parsecmap(int)
_float_ = parsec.joint(AT_MOST_ONE(dash_thing),
                       parsec.try_choice(parsec.joint(digits1, point, digits).parsecmap(join),
                                         parsec.joint(digits, point, digits1).parsecmap(join))
                      ).parsecmap(join)
float_ = _float_.parsecmap(float)
E = parsec.string('E')
times = parsec.string('*')
scientific_notation = parsec.joint((_float_ ^ _int_), E, _int_).parsecmap(join).parsecmap(float)
num = OR(scientific_notation,
         float_,  # float first so int can't capture
         int_,
         num_word)
#num = parsec.regex(r'-?(0|[1-9][0-9]*)([.][0-9]+)?([eE][+-]?[0-9]+)?')
siprefix = OR(*make_funcs(coln(0, _SIPREFS), _siplookup))
siunit = OR(*make_funcs(list(coln(0, _SIUNITS + _EXTRAS)) +
                        list(coln(1, _SIUNITS + _EXTRAS)), _silookup))
@parsec.generate
def unit_atom():
    unit = yield (siprefix + siunit).parsecmap(lambda v: (v[1], v[0])) ^ siunit.parsecmap(BOX)
    return unit

@parsec.generate
def unit_base():
    atom = yield unit_atom << spaces
    exp = yield AT_MOST_ONE(AT_MOST_ONE(exponent).parsecmap(lambda v: 'exponent') + (spaces >> int_))
    if exp:
        return atom, exp
    return atom

@parsec.generate
def unit():
    u = yield unit_base
    op = yield spaces >> AT_MOST_ONE(unit_op)
    more = yield parsec.many(spaces >> unit)
    if op and more:
        return (u, op, *more)
    elif op:
        return u, op
    elif more:
        return (u, '*', *more)  # no operator interpreted as multiplication eg kgm/s2
    else:
        return u

to = parsec.string('to')
range_indicator = (thing_accepted_as_a_dash ^ to).parsecmap(lambda v: 'range')
ph = parsec.string('pH')
P = parsec.string('P')
post_natal_day = P.parsecmap(lambda v: "'postnatal-days")
fold_prefix = parsec.ends_with(by, num).parsecmap(lambda v: "'fold")
prefix_unit = param('prefix-unit')(OR(ph, post_natal_day, fold_prefix).parsecmap(BOX))
prefix_quantity = (prefix_unit + (spaces >> num)).parsecmap(lambda v: (v[1], v[0]))
percent = parsec.string('%').parsecmap(lambda v: "'percent").parsecmap(BOX)
fold_suffix = parsec.ends_with(by.parsecmap(lambda v: "'fold"), parsec.eof() ^
                               parsec.none_of('1234567890')).parsecmap(BOX)
suffix_unit_no_space = param('unit')(EXACTLY_ONE(fold_suffix))
suffix_unit = param('unit')(percent ^ unit)
@parsec.generate
def suffix_quantity():
    v = yield num
    u = yield suffix_unit_no_space ^ (spaces >> AT_MOST_ONE(suffix_unit))
    return v, u if u else tuple()

quantity = param('quantity')(prefix_quantity ^ suffix_quantity)

@param('dilution')
@parsec.generate
def dilution_factor():
    one = yield int_
    yield colon
    times = yield int_
    return one, times

@param('dimensions')
@parsec.generate
def dimensions():
    first = yield quantity
    yield spaces >> by
    middle = yield parsec.many((spaces >> quantity) << (spaces >> by))
    last = yield spaces >> quantity
    return (first, *middle, last)

prefix_operator = plus_or_minus ^ comparison
infix_operator = OR(plus_or_minus, range_indicator, multiplication, division, exponent)
prefix_expression = prefix_operator + (spaces >> quantity)

@parsec.generate
def infix_expression():
    start = yield quantity << spaces
    op = yield infix_operator << spaces
    end = yield infix_expression ^ quantity.parsecmap(BOX)
    return (start, op, *end)

expression = param('expression')(prefix_expression ^ infix_expression)

def approximate_thing(parser):
    return EXACTLY_ONE(approx) + (spaces >> parser)

C_for_temp = param('unit')(parsec.string('C').parsecmap(lambda v: (_silookup['degrees-celcius'],)))
temp_for_biology = num + C_for_temp

@parsec.generate
def parameter_expression():
    aval = yield AT_MOST_ONE(approx << spaces)
    value = yield OR(dimensions,
                     dilution_factor,
                     temp_for_biology,
                     expression,
                     quantity,
                    )
    if aval:
        return aval, value
    return value

@parsec.Parser
def FAILED(text, index):
    return parsec.Value.success(len(text), ('parse-failure', text))

def main():
    tests = ('1 daL', "300 mOsm", "0.5 mM", "7 mM", "0.1 Hz.", "-50 pA",
             "200–500mm", "0.3%–0.5%", "1:500", "4%", "10 U/ml",
             "–20°C", "<10 mV", "–70 ± 1 mV", "30 to 150 pA",
             "310 mosmol/l", "13–16 days old", "50 x 50 um",
             "~3.5 - 6 Mohms", "pH 7.3", "17–23 d old", "10 –100",
             "250 +- 70 um", "20±11 mm", "+- 20 degrees",
             '0.1 mg kg–1', '75  mg / kg', '40x', 'x100',
             '200μm×200μm×200μm', '20--29 days', '4 °C', '10×10×10',
             '10 kg * mm^2 / s^2', '10 * 1.1 ^ 30 / 12'
            )
    weirds = ("One to 5", "100-Hz", "25 ng/ul)", "34–36°C.",
              '3*10^6 infectious particles/mL',
              '4.7 +- 0.6 x 10^7 / mm^3',  # FIXME this is ambigious? YES VERY also unit dimensionality...
              '1,850', '4C', 'three', 'Four', 'P28.5±2 days'
             )
    should_fail = ('~~~~1',
                   "(pH 7.3",
                  )
    #fun = [t.split(' ')[-1] for t in tests][:5]
    #test_unit_atom = [unit_atom(f) for f in fun]
    #test_unit = [unit(f) for f in fun]
    #test_quantity = [quantity(t) for t in tests]
    q = "'"
    test_expression = '\n'.join(f"'{t+q:<25} -> {parameter_expression(t, 0).value}" for t in tests + weirds)
    print(test_expression)
    #test_fails = [parameter_expression(t) for t in tests]
    embed()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
""" parse Paxinos atlas index formats """
from pysercomb.parsing import *

LQUOTE = '‘'
RQUOTE = '’'
COMMA = ','

lquote = COMP(LQUOTE)
rquote = COMP(RQUOTE)
comma = COMP(COMMA)

cs = JOINT(comma, space)

wsoc = OR(comment, whitespace)
wsoc1 = OR(comment, whitespace1)
word = joinstr(MANY1(NOT(whitespace1)))

def make_figalph(v):
    return RETURN(tuple(f'{v[0]}{_}' for _ in v[1]))

def flatten_alph(v):
    out = tuple(i for e in v
                for i in (e
                          if (type(e) == tuple and
                              (any(type(_) == str or type(_) == tuple for _ in e) or
                               len(e) <= 1))
                          else ((e,) if e else e)))
    return RETURN(out)

figure_range = JOINT(int_, COMPOSE(hyphen_minus, int_), join=False)
figure_alpha_list = BIND(JOINT(int_, BIND(JOINT(char,
                                                MANY(COMPOSE(comma, char))),
                                          flatten)),
                         make_figalph
                        )

figure = END(int_, OR(cs, newline, comment))

flist_atom = OR(figure_range,  # FIXME bad implementation use lookhead if possible
                figure_alpha_list,
                figure)

figures = BIND(JOINT(flist_atom,
                     MANY(COMPOSE(cs, flist_atom))),
               flatten_alph)

abrev = END(word, space)
sbrev = END(word, OR(newline, comment))
alabel = joinstr(BIND(JOINT(word,
                            BIND(MANY(JOINT(END(space,
                                                NOT(flist_atom)),
                                            word)),
                                 flatten1)),
                      flatten))
slabel = joinstr(BIND(JOINT(word,
                            BIND(MANY(JOINT(END(space,
                                                NOT(sbrev)),
                                            word)),
                                 flatten1)),
                      flatten))

los = RETVAL(LEXEME(COMP('List of Structures')), 's')
loa = RETVAL(LEXEME(COMP('List of Abbreviations')), 'a')

arec = LEXEME(JOINT(abrev, COMPOSE(space, alabel), COMPOSE(space, figures)))  # FIXME Ce has no structure name
srec = LEXEME(JOINT(slabel, COMPOSE(space, sbrev)))

def section_out(v):
    label, many, last = v
    out = label, many + (last,)
    return RETURN(out)

section = BIND(OR(JOINT(los,
                        MANY1(END(srec,
                                  NOT(OR(loa, EOF)))),
                        srec),
                  JOINT(loa,
                        MANY1(END(arec,
                                  NOT(OR(los, EOF)))),
                        arec)),
               section_out)
sections = JOINT(section, section)

def main():
    tests = dict(
    atest4_0 = 'FAKE fake fake fake 117a    ; testing',
    atest4_1 = 'CeCv central cervical nucleus 117a,b',
    atest4_2 = 'Cg1 cingulate cortex, area 1 7-24, 79-82',
    atest4_3 = 'AA anterior amygdaloid area 25, 84-87, 93-95',
    atest4_4 = 'mfba medial forebrain bundle, ‘a’ component 9-21, 81-85',


    stest4_0 = 'ventromedial hypothalamic nucleus, central part VMHC\nnext a  ; testing\n  ',
    stest4_1 = 'ventromedial hypothalamic nucleus, central part VMHC  ; testing\n  ',
    stest4_2 = 'ventromedial hypothalamic nucleus, central part VMHC',
    stest4_3 = 'olfactory ventricle (olfactory part of lateral ventricle) OV'    ,

    stest6_1 = '4th and 5th cerebellar lobules 4/5Cb',
    atest6_1 = '9a,bCb 9th cerebellar lobule, a and b 140-159, 165',
    )

    ats = [(v, arec(v)) for k, v in tests.items() if k.startswith('a')]
    sts = [(v, srec(v)) for k, v in tests.items() if k.startswith('s')]

if __name__ == '__main__':
    main()



import numbers
from itertools import chain
from pysercomb import parsing
from pysercomb.utils import coln, log
from pysercomb.parsing import *
from pysercomb.parsers.racket import racket_doc, racket_module

totally_not_degrees = 'º◦˚⁰'  # come join the merry fools!
_superscript_zero = b'\xe2\x81\xb0'  # fails to read in a racket file on windows but not here ???


def LEXEME(func):
    return COMPOSE(whitespace, SKIP(func, whitespace))


def get_quoted_list(folderpath, filename):
    with open((folderpath / filename).as_posix(), 'rt') as f:
        src = f.read()
        success, value, rest = racket_doc(src)

    if not success:
        raise SyntaxError(f'Something is wrong in {filename}. Parse output:\n{value}\n\n{rest}')

    module = racket_module(value)
    out = {k.replace('-', '_'):v for k, v in module.items()}
    return out


def fix_plurals(ql):
    # XXX hack to deal with bad parsing of plurals
    # this solution can induce nasty performance issues
    # the correct solution is to build the parser so that
    # the plural form is optional and there are shared prefixes
    # this approach is quick and dirty so we don't worry about it

    def fix(u):
        if u == 'inches':
            return 'inch'
        else:
            return u[:-1]

    for k, v in ql.items():
        more = tuple([(fix(u), u) for p, u in v if u.endswith('s') and len(u) > 2 and not u.endswith('celsius')])
        ql[k] += more


def op_order(return_value):
    order = 'plus-or-minus', 'range', '^', '/', '*', '+', '-'
    associative = '*', '+'
    commutative = '*', '+'
    def key(subtree):
        if len(subtree) == 3 and isinstance(subtree[0], str) and subtree[0].startswith('param:'):
            return 0, subtree[1]  # short sort first
        else:
            return 1, len(subtree)

    def inner(subtree):
        for op in order[::-1]:
            try:
                i = subtree.index(op)
                front = subtree[:i]
                rest = subtree[i + 1:]
                f = inner(front)
                r = inner(rest)
                if op in commutative:
                    if isinstance(f, numbers.Number):
                        lf = 0
                    else:
                        lf = len(f)

                    if isinstance(r, numbers.Number):
                        lr = 0
                    else:
                        lr = len(r)

                    if lf > lr:
                        f, r = r, f
                    elif lf == lr != 0:
                        fcase = lf == 3 and isinstance(f[0], str) and f[0].startswith('param:')
                        rcase = lr == 3 and isinstance(r[0], str) and r[0].startswith('param:')
                        if fcase and rcase:
                            if isinstance(f[1], tuple):  # number expression
                                if isinstance(r[1], tuple):
                                    l1f, l1r = len(f[1]), len(r[1])
                                    if l1f > l1r:
                                        f, r = r, f  # put longer last
                                    elif l1f == l1r:
                                        pass   # TODO check relative values?
                                else:
                                    f, r = r, f  # put tuples last

                            elif isinstance(r[1], tuple):  # number expression
                                pass  # already ordered correctly with tuple last

                            elif r[1] < f[1]:  # reorder commutative operations to be deterministic
                                f, r = r, f

                        elif rcase:  # fcase is the default so don't need to handle it
                            f, r = r, f

                if op in associative:  # assoc -> ok to move parens
                    if isinstance(f, numbers.Number) and isinstance(r, numbers.Number):
                        if f < r:
                            val = f, r
                        else:
                            val = r, f

                        return (op, *val)

                    elif isinstance(f, numbers.Number):
                        if r[0] == op:
                            # this means that the operator on f is also op
                            # and so if op commutes then we can reorder
                            if op in commutative:
                                val = sorted((f, *r[1:]))
                            else:
                                val = f, *r[1:]

                        else:
                            val = f, r

                    elif isinstance(r, numbers.Number):
                        if f[0] == op:
                            if op in commutative:
                                val = sorted((f[1:], r))
                            else:
                                val = *f[1:], r
                        else:
                            val = r, f

                    elif f[0] == op and r[0] == op:
                        val = *f[1:], *r[1:]
                        if op in commutative:
                            val = sorted(val, key=key)

                    elif f[0] == op:
                        if op in commutative:
                            val = *sorted(f[1:], key=key), r
                        else:
                            val = *f[1:], r

                    elif r[0] == op:
                        # this means that the operator on f is also op
                        # and so if op commutes then we can reorder
                        # I think this works because it implies that f[0] is absent??!
                        if op in commutative:
                            val = sorted((f, *r[1:]), key=key)
                        else:
                            val = f, *r[1:]

                    else:
                        val = f, r
                else:
                    val = f, r

                return (op, *val)
            except ValueError:
                # can't index because we are at the bottom
                # FIXME this has ambiguous semantics
                continue

        return subtree[0]

    log.debug(return_value)
    lisped = inner(return_value)
    log.debug(lisped)
    if lisped != return_value:
        return RETURN((lisped,))
    else:
        return RETURN(return_value)


def param(prefix_name):
    name = 'param:' + prefix_name
    def add_function_type(v):
        return RETURN((name, *v))
    def paramed(parser_func):
        return BIND(parser_func, add_function_type)
    return paramed


debug = param('debug')

# basic tokens and operators
_plus_or_minus = '±'  # b'\xc2\xb1' '\u00b1'
plus_or_minus_symbol = COMP(_plus_or_minus)  # NOTE range and +- are interconvertable...
plus_or_minus_pair = COMP('+-')  # yes that is an b'\x2d'
plus_over_minus = COMP('+/-')
plus_or_minus = RETVAL(OR(plus_or_minus_symbol, plus_or_minus_pair, plus_over_minus), 'plus-or-minus')

addition = COMP('+')
subtraction = dash_thing
division = OR(COMP('/'), RETVAL(COMP('per'), '/'))  # also appears as x a/an time-unit, or x per something XXX TODO log normalizations
multiplication = RETVAL(OR(COMP('*'), by, parsing.dot), '*')  # ok as long as dimensions are specced with units 1 mm x 1 mm  # XXX FIXME by now includes X
# there is actually no way to tell the difference between
# 10 x 20 mm as 10 mm x 20 mm and 10 x 20 mm = 200 mm
math_op = OR(addition, subtraction, division, multiplication, exponent)  # FIXME subtraction is going to be a pain
unit_op = OR(division, multiplication)
lt = COMP('<')
_less_than_or_equal_to = '≤'  # b'\xe2\x89\xa4' '\u2264'
lte_symbol = COMP(_less_than_or_equal_to)
lte_pair = COMP('<=')
lte = RETVAL(OR(lte_pair, lte_symbol), '<=')
gt = COMP('>')
_greater_than_or_equal_to = '≥'  # b'\xe2\x89\xa5' '\u2265'
gte_symbol = COMP(_greater_than_or_equal_to)
gte_pair = COMP('>=')
gte = RETVAL(OR(gte_pair, gte_symbol), '>=')
comparison = OR(lte, gte, lt, gt)
approx = RETVAL(COMP('~'), 'approximately')

def get_unit_dicts(units_path):
    return [get_quoted_list(units_path, _) for _ in
            ('si-prefixes-data.rkt',
             'si-prefixes-exp-data.rkt',
             'si-units-data.rkt',
             'si-units-extras.rkt',
             'units-dimensionless.rkt',
             'imperial-units-data.rkt')]


DEGREES_UNDERLINE = b'\xc2\xba'.decode()  # º sometimes pdfs misencode these
DEGREES_FEAR = b'\xe2\x97\xa6'  # this thing is scary and I have no idea what it is or why it wont change color ◦
RING_ABOVE = b'\xcb\x9a'  # u02da apparently this thing can creep in as a degrees symbol sometimes as well ??!?

def hms(h, m, s):
    """ hours minutes seconds to seconds """
    return h * 60 * 60 + m * 60 + s

# units
def make_unit_parser(units_path=None, dicts=None):
    if units_path is None and dicts is None:
        raise TypeError('must have units_path or dicts')
    if dicts is None:
        dicts = get_unit_dicts(units_path)

    gs = globals()
    for dict_ in dicts:
        if [k for k in ['units_si', 'units_extra', 'units_imp'] if k in dict_]:
            fix_plurals(dict_)
        gs.update(dict_)

    _silookup = {k: ('quote', v)
                 for k, v in chain(units_si,
                                   units_extra,
                                   ([v, v] for k, v in units_si),
                                   ([v, v] for k, v in units_extra))}
    _implookup = {k: ('quote', v)  # imperial separate because they don't support prefixes
                  for k, v in chain(units_imp,
                                    units_dimensionless,
                                    ([v, v] for k, v in units_imp),
                                    ([v, v] for k, v in units_dimensionless))}
    _siplookup = {k: ('quote', v) for k, v in chain(prefixes_si,
                                                    ([v, v] for k, v in prefixes_si))}

    siprefix = OR(*make_funcs(chain(coln(0, prefixes_si),
                                    coln(1, prefixes_si)),
                              _siplookup))
    siunit = OR(*make_funcs(chain(coln(0, units_si + units_extra), # need both here to avoid collisions in unit_atom slower but worth it?
                                  coln(1, units_si + units_extra)),
                            _silookup))
    impunit = OR(*make_funcs(chain(coln(0, units_imp + units_dimensionless),
                                   coln(1, units_imp + units_dimensionless)),
                             _implookup))

    def parenthized(func): return COMPOSE(LEXEME(COMP('(')), SKIP(func, LEXEME(COMP(')'))))
    def parOR(func): return OR(func, parenthized(func))

    to = COMP('to')
    dtod = COMP('-to-')  # not matching ...
    range_indicator = RETVAL(OR(dtod, thing_accepted_as_a_dash, to), 'range')
    # FIXME range vs minus ...
    infix_operator = OR(plus_or_minus, range_indicator, math_op)  # colon? doesn't really operate on quantities, note that * and / do not interfere with the unit parsing because that takes precedence

    def num_expression(thing): return OR(param('expr')(parOR(_num_expression)), parOR(OR(num_commas, num)))(thing)
    def num_expression_inner(thing): return OR(parOR(_num_expression), OR(num_commas, num))(thing)

    _dig = BIND(parsing.TIMES(digit, 1, 3), lambda i: RETURN(''.join(i)))
    _3dig = BIND(parsing.TIMES(digit, 3, 3), lambda i: RETURN(''.join(i)))
    num_commas = BIND(JOINT(_dig, MANY1(COMPOSE(COMP(','), _3dig)), join=True),
                      lambda i: RETURN(int(''.join(i))))

    num_thing = OR(parOR(OR(num_commas, num)), BIND(parenthized(num_expression_inner), flatten1))
    num_suffix = JOINT(COMPOSE(spaces, infix_operator),
                       COMPOSE(spaces, num_thing))

    _num_expression = BIND(BIND(JOINT(num_thing,
                                      BIND(MANY1(num_suffix),
                                           flatten1)),
                                flatten),
                           op_order)

    # biology specific extensions
    _C_for_temp = COMP('C')
    C_for_temp = RETVAL(_C_for_temp, BOX(_silookup['degrees-celsius']))
    temp_for_biology = JOINT(num, C_for_temp, join=False)  # XXX not used
    mmHg = BIND(OR(COMP('mmHg'), COMP('mm Hg')), (lambda _: RETURN((('quote', 'millimeters-hg'),))))  # FIXME move to data files
    RCF = BIND(OR(COMP('RCF'), COMP('xg')), (lambda _: RETURN((_implookup['RCF'],))))  # since ronna R broke all the things  XXX xg lacks priority due to fold being in suffix unit no space
    dpi = BIND(COMP('dpi'), (lambda _: RETURN((('quote', 'pixels-per-inch'),))))  # usually uppercase
    micron = BIND(COMP('micron'), (lambda _: RETURN((('quote', 'micro'), ('quote', 'meters')))))
    def sigh(p): return parsing.oper(p, lambda s: s in totally_not_degrees)
    degc_mess = BIND(OR(COMP('degrees Celsius'),
                        COMP('degrees C'),
                        COMP('o celcius'),
                        COMP('oC'),
                        JOINT(sigh, _C_for_temp)),
                     (lambda _: RETURN((_silookup['~oC'],))))

    manual_unit = OR(mmHg, RCF, degc_mess, micron, dpi)
    unit_atom = param('unit')(BIND(OR(manual_unit,
                                      JOINT(siprefix, siunit, join=False),
                                      BIND(impunit, RETBOX),  # FIXME R RCF collision
                                      BIND(siunit, RETBOX)),  # merge just units?
                                   FLOP))

    maybe_exponent = LEXEME(AT_MOST_ONE(exponent))  # XXX must be AT_MOST_ONE, but issues with ranges, see test_exp_vs_range

    exp_short = BIND(JOINT(SKIP(COMPOSE(spaces, unit_atom),  # XXX TODO FIXME need unit_atom_not_dimensionless
                                COMPOSE(spaces, maybe_exponent)),
                           int_),
                     lambda v: RETURN(('^', *v)))
    exp_no_op = (BIND(exp_short, lambda v: RETURN(('*', v))))  # FIXME no preceeded by an operator ...
    unit_thing = OR(exp_short, unit_atom)
    def unit_expr(thing): return parOR(unit_expr_atom)(thing)
    def unit_par(thing): return parenthized(unit_expr_atom)(thing)
    #unit_implicit_count_ratio = BIND(LEXEME(division), lambda v: RETBOX((v[0], unit("count")[1], *v[1:])))
    def UICR(func):
        return BIND(JOINT(LEXEME(division), func),
                    lambda v: RETBOX((v[0], unit("count")[1], *v[1:])))
    #unit_op_prefix = unit_implicit_count_ratio
    unit_suffix = OR(JOINT(COMPOSE(spaces, unit_op),
                           COMPOSE(spaces, OR(unit_thing, BIND(unit_expr, flatten1)))),
                     COMPOSE(spaces, exp_no_op))
    # I think the issue is cases like 1 + (2 * 3) - 4 ...
    unit_expr_atom = BIND(BIND(JOINT(OR(UICR(unit_thing), #JOINT(unit_op_prefix, unit_thing),
                                        unit_thing,
                                        BIND(unit_par, flatten1)),
                                     BIND(MANY1(unit_suffix),
                                          flatten1)),
                               flatten),
                          op_order)

    unit_expression = param('unit-expr')(OR(unit_expr,
                                            BIND(exp_short, RETBOX),
                                            UICR(unit_expr),
                                            UICR(exp_short)))

    unit = OR(unit_expression, unit_atom)

    unit_starts_with_dash = COMPOSE(dash_thing, unit)  # derp
    #unit_implicit_count_ratio = param('unit-expr')(_unit_implicit_count_ratio)

    def plus_or_minus_thing(thing): return JOINT(plus_or_minus, COMPOSE(spaces, thing), join=False)

    def range_thing(func): return JOINT(func, COMPOSE(spaces, range_indicator), COMPOSE(spaces, func))

    # XXX FIXME loss of representation issues are possible here
    # the solution is to represent the value as a sum I think
    _dur_digits = BIND(MANY1(digit), lambda v: RETURN(''.join(v)))
    _dur_number = OR(BIND(JOINT(_dur_digits, RETVAL(OR(comma, point), '.'), _dur_digits, join=True),
                          # FIXME precision etc.
                          lambda v: RETURN(float(''.join(v)))),
                     # FIXME -> represent
                     BIND(_dur_digits, lambda v: RETURN(int(v))))
    _, _dur_date, _, _dur_time = (
        (_dur_P,),
        (_dur_Y, _dur_M, _dur_D),
        (_dur_T,),
        (_dur_H, _dur_M, _dur_S)
    ) = [[RETVAL(COMP(l), val) for l, val in zip(ls, vals)]
         for ls, vals in zip(('P', 'YMD', 'T', 'HMS'),
                       ((None,),
                        [('quote', u) for u in ('years', 'months', 'days')],
                        (None,),
                        [('quote', u) for u in ('hours', 'minutes', 'seconds')]))]
    # technically these could be parsed in any order because we know the type of
    # the individual duration, however there are some weird impliciations about
    # the ordering, depending on which section you count first
    #  012   01   02   12   0   1   2
    # 'YMD' 'YM' 'YD' 'MD' 'Y' 'M' 'D'
    # 'HMS' 'HM' 'HS' 'MS' 'H' 'M' 'S'
    _inds = [[0, 1, 2], [0, 1], [0, 2], [1 ,2], [0], [1], [2]]
    combs = _c_date, _c_time = [[[param('quantity')
                                  (ANDTHEN(_dur_number,
                                           param('unit')
                                           (BIND(elems[index], RETBOX))))
                                  for index in indexes]
                                 for indexes in _inds]
                                for elems in (_dur_date, _dur_time)]
    _p_date, _p_time = [OR(*[JOINT(*parts, join=False)
                             for parts in comb])
                        for comb in combs]
    _tp_time = COMPOSE(_dur_T, _p_time)
    _p_datetime = JOINT(_p_date, _tp_time, join=True)
    iso8601duration = COMPOSE(
        _dur_P,
        # these sort into different functions because durations that
        # include dates are invariants while time alone is a parameter
        # this helps with equality checking among other things
        OR(BIND(_tp_time,
                lambda v: RETURN(('iso8601-duration-time', *v))),
           BIND(_p_datetime,
                lambda v: RETURN(('iso8601-duration-datetime', *v))),
           BIND(_p_date,
                lambda v: RETURN(('iso8601-duration-date', *v))),))

    pH = RETVAL(COMP('pH'), BOX("'pH"))
    post_natal_day = RETVAL(COMP('P'), BOX(('quote', 'postnatal-days')))  # FIXME note that in our unit hierarchy this is a subclass of days
    _fold_prefix = END(by, num)
    fold_prefix = RETVAL(_fold_prefix, BOX(('quote', 'fold')))

    prefix_unit = param('prefix-unit')(OR(pH, post_natal_day, fold_prefix))
    _prefix_quantity = JOINT(prefix_unit, COMPOSE(spaces, num_expression))  # OR(JOINT(fold, num))
    prefix_quantity = BIND(_prefix_quantity, FLOP)

    # num only prefix quantities
    _prefix_quantity_num_only = JOINT(prefix_unit, COMPOSE(spaces, num))  # OR(JOINT(fold, num))
    prefix_quantity_num_only = BIND(_prefix_quantity, FLOP)

    fold_suffix = RETVAL(END(by, noneof('0123456789')), BOX(('quote', 'fold')))  # NOT(num) required to prevent issue with dimensions
    suffix_unit = unit #OR(unit_implicit_count_ratio, unit)
    suffix_unit_no_space = OR(param('unit')(OR(EXACTLY_ONE(fold_suffix), C_for_temp)), unit_starts_with_dash)  # FIXME this is really bad :/ and breaks dimensions...
    suffix_quantity = JOINT(num_expression,
                            OR(suffix_unit_no_space,
                               COMPOSE(spaces,
                                       AT_MOST_ONE(suffix_unit, fail=False))))  # this catches the num by itself and leaves a blank unit
    def infix_par(thing): return parenthized(infix_expression)(thing)
    quantity = param('quantity')(OR(prefix_quantity, suffix_quantity))

    # quantity expressions that require a trailing unit
    suffix_quantity_with_unit = JOINT(num_expression,
                                      OR(suffix_unit_no_space,
                                         COMPOSE(spaces,
                                                 suffix_unit)))
    quantity_with_unit = param('quantity')(OR(prefix_quantity, suffix_quantity_with_unit))

    # quantity expressions that require no numerical expressions, only numbers
    suffix_quantity_num_only = JOINT(OR(num_commas, num),
                                     OR(suffix_unit_no_space,
                                        COMPOSE(spaces,
                                                AT_MOST_ONE(suffix_unit, fail=False))))
    quantity_num_only = param('quantity')(OR(prefix_quantity_num_only, suffix_quantity_num_only))

    _hmsret = lambda v: param('quantity')(JOINT(RETURN(hms(*v)),
                                                param('unit')(RETURN((('quote', 'seconds'),)))))
    duration_hms = BIND(JOINT(SKIP(int_, colon), SKIP(int_, colon), int_, join=False), _hmsret)  # FIXME loss of repr
    #duration_hm = BIND(JOINT(SKIP(int_, colon), int_, join=False), _hmsret)  # TODO FIXME loss of repr
    ratio = param('ratio')(JOINT(SKIP(int_, colon), int_, join=False))  # dilution starts with 1 but is an aspect
    sq = COMPOSE(spaces, quantity)
    sby = COMPOSE(spaces, by)
    dimensions = param('dimensions')(BIND(JOINT(quantity_with_unit,
                                                MANY1(COMPOSE(COMPOSE(spaces, SKIP(by, spaces)),
                                                              END(quantity, NOT(exponent))))),  # catch A x B^C
                                          flatten))

    # use this in situations where A x B implies dimensions rather than multiplication
    # this is no longer the default assumption for the parser
    dimensions_no_math = param('dimensions')(BIND(JOINT(quantity_num_only,
                                                        MANY1(COMPOSE(COMPOSE(spaces, SKIP(by, spaces)),
                                                                      END(quantity, NOT(exponent))))),
                                                  flatten))
    prefix_operator = OR(approx, plus_or_minus, comparison)  # FIXME approx is not really an operator
    # TODO when parsing approx actually has higher affinity for the
    # number than the units, because it modifies the certainty about
    # the magnitude independent of the units and thus should probably
    # parse as (quantity (~ 10)) instead of (~ (quantity 10))
    def infix_expr(thing): return parOR(infix_expression)(thing)
    infix_suffix = JOINT(COMPOSE(spaces, infix_operator),
                         COMPOSE(spaces, OR(quantity, BIND(infix_expr, flatten1))))
    infix_expression = BIND(BIND(JOINT(OR(quantity, BIND(infix_par, flatten1)),
                                       BIND(MANY1(infix_suffix),
                                            flatten1)),
                                 flatten),
                            op_order)
    prefix_expression = BIND(BIND(JOINT(prefix_operator,
                                        COMPOSE(spaces,
                                                OR(infix_expression,
                                                   BIND(quantity, RETBOX)))),
                                  flatten), RETBOX)
    expression = param('expr')(OR(prefix_expression, infix_expression))  # FIXME this doesn't work if you have prefix -> infix are there cases that can happen?

    boo = param('bool')(BIND(boolean, lambda v: RETBOX('#t') if v else RETBOX('#f')))

    #def approximate_thing(thing): return JOINT(EXACTLY_ONE(approx), COMPOSE(spaces, thing), join=False)

    def FAILURE(p):
        # NOTE we do not return (True, (p,), p) because FAILURE
        # should only be used in the last slot, nothing should be
        # parsing after this in the same chain
        return param('parse-failure')(lambda null: (True, (p,), ''))(p)

    # TODO objective specifications...
    components = OR(dimensions,
                    duration_hms,  # must come first otherwise ratio will always match first
                    ratio,
                    iso8601duration,  # FIXME it should be possible to compose this
                    expression,
                    quantity,
                    boo,
                    FAILURE)
    #approx_comp = approximate_thing(components)


    parameter_expression = LEXEME(components)  # OR(approx_comp, components)

    debug_dict = {'infix_expression': infix_expression,
                  'prefix_expression': prefix_expression,
                  'unit_expr_atom': unit_expr_atom,
                  'unit_expr': unit_expr,
                  'unit_expression': unit_expression,
                  'num_expression': num_expression,
                  'suffix_unit': suffix_unit,
                  'exp_short': exp_short,
                  'unit_thing': unit_thing,
                  'expression': expression,
                  'num_thing': num_thing,
                  'dimensions_no_math': dimensions_no_math,  # TODO
                  'iso8601duration': iso8601duration,
                  'siprefix': siprefix,
                 }

    return parameter_expression, quantity, unit, unit_atom, debug_dict


def parse_for_tests(parameter_expression=None):

    tests = ('1 daL', "300 mOsm", "0.5 mM", "7 mM", "0.1 Hz.", "-50 pA",
             "200–500mm", "0.3%–0.5%", "1:500", "4%", "10 U/ml",
             "–20°C", "<10 mV", "–70 ± 1 mV", "30 to 150 pA",
             "310 mosmol/l", "13–16 days old", "50 x 50 um",
             "~3.5 - 6 Mohms", "pH 7.3", "17–23 d old", "10 –100",
             "250 +- 70 um", "20±11 mm", "+- 20 degrees",
             '0.1 mg kg–1', '75  mg / kg', '40x', 'x100',
             '200μm×200μm×200μm', '20--29 days', '4 °C', '10×10×10',
             '10 kg * mm^2 / s^2', '10 * 1.1 ^ 30 / 12',
             '120 +- 8 * 10 ^ 6 MR / kg * s2 * 20',
             '1 * 2 * 3 * 4 * 5', '1 + 2 + 3 + 4 + 5',
             '10lbs', '11 lbs',
             'TRUE', 'fAlsE', '#t', '#f',
             '10 \u00B5L', '10 \u03BCL',  # micro issues
             '1 s-1', '1 h-1', '1 mg kg-1 h-1',
            )

    prefix_expr_tests = ('<1', '~3.5 - 6 Mohms',)

    weirds = ("One to 5", "100-Hz", "25 ng/ul)", "34–36°C.",
              '3*10^6 infectious particles/mL',
              '4.7 +- 0.6 x 10^7 / mm^3',  # FIXME this is ambigious? YES VERY also unit dimensionality...
              '1,850', '4C', 'three', 'Four', 'P28.5±2 days',
              '12-V', '10-mL', '+1', 'Forty seconds', '10s - 100h'
             )
    should_fail = ('~~~~1',
                   "(pH 7.3",
                  )

    from pathlib import Path
    from protcur.config import __units_test_params__ as test_params
    with open(test_params, 'rt') as f:
        success, value, rest = racket_doc(f.read())

    module = racket_module(value)
    param_test_strings = module['param-test-strings']

    _all = tests + prefix_expr_tests + weirds + should_fail + param_test_strings
    if parameter_expression:
        from pysercomb.pyr import units as pyru
        up = pyru.UnitsParser
        parsed = [up(t) for t in _all]
    else:
        parsed = None

    return tests, prefix_expr_tests, weirds, should_fail, param_test_strings, _all, parsed


def main():
    import pprint
    from time import time
    from pathlib import Path
    from pysercomb.pyr.units import SExpr
    from protcur.config import __units_folder__ as units_path

    try:
        from desc.prof import profile_me
    except ImportError:
        profile_me = lambda f: f

    (parameter_expression, quantity, unit,
     unit_atom, debug_dict) = make_unit_parser(units_path)

    (tests, prefix_expr_tests, weirds, should_fail, param_test_strings, _all,
     parsed) = parse_for_tests()

    test_all = []

    pid = []
    def timeit():
        for t in param_test_strings:
            success = False
            t2 = t
            while t2 and not success:
                _, v, rest = parameter_expression(t2)
                success = v[0] != 'param:parse-failure'
                if not success:
                    t2 = t2[1:]
            if not success:
                rest = t
            else:
                pid.append(SExpr(v))

            test_all.append((success, v, rest))
    start = time()
    timeit()
    stop = time()
    print('BAD TIME', stop - start)
    print_issues = {i:h for i, h in enumerate(pid)} 
    # pprint.pprint(print_issues)  # this works correctly

    q = "'"
    fun = [t.split(' ')[-1] for t in tests][:5]
    test_unit_atom = [unit_atom(f) for f in fun]
    test_unit = [unit(f) for f in fun]
    test_quantity = [quantity(t) for t in tests]
    test_expression = [parameter_expression(t) for t in tests + prefix_expr_tests + weirds]
    test_expression2 = '\n'.join(sorted((f"'{t+q:<25} -> {parameter_expression(t)[1]}"
                                         for t in tests + weirds), key=lambda v: v[25:]))
    print(test_expression2)
    test_fails = [parameter_expression(t) for t in tests]
    if __name__ == '__main__':
        from IPython import embed
        embed()


if __name__ == '__main__':
    main()



from pathlib import Path
from protcur.config import __units_folder__ as units_folder
from pysercomb import parsing
from pysercomb.parsers import units

(parameter_expression, quantity, unit, *_,
 debug_dict) = units.make_unit_parser(units_folder)

*_, test_all, parsed = units.parse_for_tests(parameter_expression)

# evil
_gs = globals()
_gs.update(debug_dict)




import pprint
import unittest
import pytest
import rdflib
from pysercomb import exceptions as exc
from pysercomb.parsers.units import DEGREES_FEAR
from pysercomb.pyr.units import ParamParser, SExpr, Expr, UnitsParser
from pysercomb.pyr import units as pyru
from .common import *

class TestPint(unittest.TestCase):
    def test_percent(self):
        from pysercomb.pyr.units import ur
        one = ur.parse_expression('10%')
        #print(one)
        two = ur.parse_units('percent')
        #print(two)
        #three = ur.parse_units('%')  # a bare percent conflicts with python mod
        #print(three)

    @pytest.mark.skip('Expected to fail since the pint parser is simpler than ours.')
    def test_all(self):
        from pint import UnitRegistry
        bads = []
        fails = []
        for expr in test_all:
            u = None
            try:
                u = UnitsParser(expr)
                out  = u.asPython()
            except exc.BadNotationError:
                pass
            except exc.ParseFailure as e:
                fails.append((expr, e, u))
            except BaseException as e:
                raise e
                bads.append((expr, e, u))

        #assert not bads, '\n'.join(str(b) for b in bads)
        assert not fails, '\n'.join(str(f) for f in fails)




from pysercomb.parsing import *

# racket
def exp(p):
    return _exp(p)

comment = COMPOSE(whitespace,
                  COMPOSE(COMP(';'),
                          MANY(NOT(newline))))  # leave the newline intact
def LEXEME(func):
    return COMPOSE(OR(SKIP(MANY1(comment), whitespace), whitespace),
                   SKIP(func, OR(MANY1(comment), whitespace)))
open_paren = LEXEME(COMP('('))
close_paren = LEXEME(COMP(')'))
quote_symbol = COMP("'")
double_quote_symbol = COMP('"')
_string_escape = RETVAL(COMP('\\"'), '"')
_string = COMPOSE(double_quote_symbol,
                  SKIP(MANY(OR(_string_escape, NOT(double_quote_symbol))),
                       # string escape has to go first so the \ arent eaten
                       double_quote_symbol))  # TODO escape
string = LEXEME(joinstr(_string))

# #<< everything until newline must be matched on its own line exactly
# #<< match\n ------- \n match\n
# also BOF isn't quite right ... it just can't be preceeded by non-whitespace
#here_string = COMPOSE(OR(whitespace, BOF), COMP("#<<"), pattern .+ COMPOSE(newline, pattern, newline))
_here_pattern = lambda v: (COMPOSE(newline,
                                   COMPOSE(COMP(v), newline))
                           if v else
                           newline)
_here_string = BIND(SKIP(COMPOSE(COMPOSE(whitespace, COMP("#<<")),
                                 joinstr(MANY(NOT(newline)))),
                         newline),
                    lambda v: SKIP(MANY(NOT(_here_pattern(v))),
                                   _here_pattern(v)))
#here_string = BIND(joinstr(_here_string), lambda v: RETURN(dumps(v)))
here_string = joinstr(_here_string)  # FIXME how to distingish from other strings?

# XXX I don't know if we have a stack in here to store the pattern ...
# actually it is probably easier in that we can just create a new combinator
# for use inside the function itself ... need to review
symbol = OR(char, digit, COMP('-'), COMP('_'), colon, COMP('*'),
            NOT(OR(COMP('('),
                   COMP(')'),
                   quote_symbol,
                   double_quote_symbol,
                   COMP(';'),
                   whitespace1,
                   point,
                   EOF)))
empty_list = RETVAL(COMP("'()"), None)
true = RETVAL(COMP("#t"), True)
false = RETVAL(COMP("#f"), False)
num_literal = END(OR(scientific_notation, float_, int_), NOT(symbol))  # FIXME HACK
cons_pair = COMPOSE(open_paren, JOINT(SKIP(exp, point), SKIP(exp, close_paren)))
literal = OR(num_literal, string, here_string, true, false, empty_list)
atom = joinstr(MANY1(symbol))  # not quite right due to numbers
identifier = LEXEME(atom)

# Include the explicit quote since we transform everything into strings
# since python has no symbols. Eval rules should not be applied here.
# They should be applied later in a second step to obtain the python ir.
quote = LEXEME(BIND(COMPOSE(quote_symbol, exp), lambda v: RETURN(('quote', v))))
def sexp(p):
    return sexp_inner(p)
# It is consisten with racket behavior for unquoted cons pairs e.g. (1 . 2)
# to read. They error during #%app since they are not quoted
_exp = LEXEME(OR(literal, identifier, quote, sexp, cons_pair))
sexp_inner = COMPOSE(open_paren, SKIP(MANY(exp), close_paren))  # not MANY1 because () is not a parse error
lang_line = COMPOSE(COMP('#lang'), SKIP(COMPOSE(whitespace1, MANY1(atom)), COMP('\n')))
module_from_lang = BIND(ANDTHEN(EXACTLY_ONE(lang_line),
                                BIND(MANY(exp), lambda v: RETURN(('module-begin', *v)))),
                        lambda v: RETURN(('module-unexp', None, *v)))
#racket_doc = OR(COMPOSE(EXACTLY_ONE(lang_line), MANY(exp)), MANY1(exp))
racket_doc = OR(module_from_lang, MANY1(exp))
tag_doc = SKIP(JOINT(COMPOSE(open_paren,
                             COMP('tag-doc')),
                     quote,
                     quote,
                     string,),
               close_paren)
tag_docs = MANY1(tag_doc)


def racket_module(value):
    """ a very minimal racket -> python interpreter """
    _awaiting = object()
    _provide_expect = object()
    def car(tup): return tup[0]
    def cdr(tup): return tup[1:]
    def eval(env, tup):
        last = None
        for value in tup:
            if isinstance(value, tuple):
                if car(value) == 'module-unexp':
                    _, file, lang, rest = value
                    return eval(None, rest)
                elif car(value) == 'module-begin':
                    env = {_provide_expect: set()}
                    eval(env, cdr(value))
                    provide_expect = env.pop(_provide_expect)
                    out = {name:value for name, value in env.items() if name in provide_expect}
                    missing = provide_expect - set(out)
                    if missing:
                        raise ValueError(f'not provided {missing}')

                    return out
                elif car(value) == 'provide':
                    for v in cdr(value):  # FIXME too simple
                        env[_provide_expect].add(v)
                    last = None
                elif car(value) == 'define':
                    if isinstance(value[2], tuple) and len(value[2:]) > 1:
                        raise NotImplementedError('havent implemented functions yet')

                    env[value[1]] = eval(env, value[2:])
                elif car(value) == 'quote':
                    rest = value[1]
                    return rest

                last = None
            else:
                last = value

        return last

    return eval(None, value)


def main():
    import os
    print(identifier('hello world!'))
    print(quote('\'hello world!'))
    print(quote('\'"hello" world!'))
    print(string('"ARE YOU KIDDING ME \n NO???"'))
    print(tag_doc('(tag-doc \'tag "a b c")\n'))
    with open(os.path.expanduser('~/git/protc/protc-tags.rkt'), 'rt') as f:
        text = f.read()
    with open(os.path.expanduser('~/git/protc/protc-lib/protc/units/si-units-extras.rkt'), 'rt') as f:
        text2 = f.read()
    td = tag_docs(text)
    e = racket_doc(text2)


if __name__ == '__main__':
    main()



from functools import wraps
import logging


def makeSimpleLogger(name, level=logging.INFO):
    # TODO use extra ...
    logger = logging.getLogger(name)
    logger.setLevel(level)
    ch = logging.StreamHandler()  # FileHander goes to disk
    fmt = ('[%(asctime)s] - %(levelname)8s - '
           '%(name)14s - '
           '%(filename)16s:%(lineno)-4d - '
           '%(message)s')
    formatter = logging.Formatter(fmt)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


log = makeSimpleLogger('pysercomb')
logd = log.getChild('data')


def coln(n, iterable):
    """ Return an iterator on the nth column. """
    for rec in iterable:
        yield rec[n]


def express(generator_function):
    @wraps(generator_function)
    def tuplify(*args, **kwargs):
        return tuple(generator_function(*args, **kwargs))

    return tuplify



#!/usr/bin/env python3

import os
from .utils import log

infinity = 99999  # you know it baby, if we go this deep we will get recursion errors
__script_folder__ = os.path.dirname(os.path.realpath(__file__))

__all__ = [
    'ANDTHEN',
    'AT_MOST_ONE',
    'BIND',
    'BOX',
    'COMP',
    'COMPOSE',
    'END',
    'EOF',
    'EXACTLY_ONE',
    'FLOP',
    'JOINT',
    'MANY',
    'MANY1',
    'NOT',
    'OR',
    'RETURN',
    'RETURNBOX',
    'RETVAL',
    'RETBOX',
    'SKIP',
    'NOOP',

    'noneof',

    'make_funcs',
    'joinstr',
    'flatten',
    'flatten1',
    'noop',
    'breakf',
    'BREAK',

    'spaces',
    'newline',
    'whitespace',
    'whitespace1',

    'dash_thing',
    'thing_accepted_as_a_dash',

    'by',
    'char',
    'colon',
    'digit',
    'comma',
    'point',
    'exponent',
    'scientific_notation',
    'float_',
    'int_',
    'num',
    'boolean',
]


# helper types

class ScientificNotation(float):
    """ type for perserving the original format of scientific notation """
    def __new__(cls, value):
        self = super().__new__(cls, value)
        self.__original = value
        return self

    def __repr__(self):
        return self.__original


# utiltiy

def wdebug(func):
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            breakpoint()
            raise e

    return wrapped

# combinators

def RETURN(v):
    def return_(p):
        return True, v, p
    return return_

def OR(*funcs):
    def or_(p):
        for f in funcs:
            success, v, rest = f(p)
            if success:
                return success, v, rest
        return False, None, p  # explicit values since different funcs will have parsed to different depths
    return or_

def TIMES(func, min_, max_=None, fail=True):
    if not max_ and not min_:
        raise TypeError('If you want something zero times use NOT')
    def times(p):
        matches = []
        for i in range(min_):
            success, v, rest = func(p)
            if not success:
                return success, None, p
            else:
                matches.append(v)
                p = rest
        if max_ is not None:
            for i in range(max_ - min_):
                try:
                    success, v, rest = func(p)
                except EOFError:  # NOT(anything) matches EOF
                    log.debug(('caught EOF in TIMES', True, matches, p))
                    return True, tuple(matches), p
                if not success:
                    return True, tuple(matches), p
                else:
                    matches.append(v)
                    p = rest
            if fail:
                success, v, rest = func(p)
                if success:
                    return False, None, p
                else:
                    success = True
                    rest = p

        return success, tuple(matches), rest
    return times

def MANY(func):
    return TIMES(func, 0, infinity)

def MANY1(func):
    return TIMES(func, 1, infinity)

def ANDTHEN(func1, func2):
    """ match func1 func2 | rest
        (func1 func2) -> """
    def andthen(p):
        success, v1, rest = func1(p)
        if success:
            p2 = rest
            success, v2, rest = func2(p2)
            if success:
                return success, (v1, v2), rest
        return success, None, p
    return andthen

def JOINT(*funcs, join=False):  # FIXME something is up with multichar tokens when join=True
    def joint(p):
        matches = []
        rest = p
        for func in funcs:
            success, v, rest = func(rest)
            if not success:
                return success, None, p
            else:
                if join and hasattr(v, '__iter__'):
                    matches.extend(v)
                else:
                    matches.append(v)
        return success, tuple(matches), rest 
    return joint

def JOINT_OR_FIRST(*funcs, join=False):  # TODO improve performance in cases where there is a base that should return true even when the rest fails
    pass

def COMPOSE(func1, func2):
    """ match func1 func2 | rest
        func1 -x
        func2 -> value """
    def compose(p):
        success, v, rest = func1(p)
        if success:
            success, v, rest = func2(rest)
            if success:
                return success, v, rest
        return success, None, p
    return compose

def BIND(parser, func2):
    def bind(p):
        success, v , rest = parser(p)
        if success:
            parser2 = func2(v)
            success, v, rest = parser2(rest)
            if success:
                return success, v, rest
        return False, v, p
    return bind

def NOT(func):
    def not_(p):
        success, v, rest = func(p)
        if success:
            return False, v, rest
        else:
            if p:
                return True, p[0], p[1:]
            else:
                raise EOFError(f'Unhandled EOF received by NOT in {func} {func.__name__}.\n'
                               'You may have a MANY->NOT case.')
                log.debug('WAT')
                return True, None, p
    return not_

def END(func1, func2):
    """ this is a form of lookahead """
    def end_(p):
        success, v, rest = func1(p)
        if not success:
            return success, v, rest
        try:
            success2, v2, rest2 = func2(rest)
        except EOFError:  # NOT will match EOF
            return success, v, rest
        if success2:
            return success, v, rest
        else:
            return success2, v2, rest2
    return end_

def SKIP(func1, func2):
    """ match func1 func2 | rest
        func1 -> value
        func2 -x """
    def skip(p):
        success, v, rest = func1(p)
        if not success:
            return success, v, rest
        success2, v2, rest2 = func2(rest)
        if success2:
            return success2, v, rest2
        else:
            return success2, v2, rest2
    return skip


def NOOP(func):
    return (lambda p: (True, None, p))

_dobreak = False
breakf = (lambda p: (False, breakpoint() if _dobreak else None, p))
def BREAK(func):  # break on entry
    def f(p):
        if _dobreak:
            breakpoint()
        return func(p)
    return f

def comp(p, val, lv):
    if p:
        v = p[:lv]  # warning: this will produce order dependencies when you spec the parser
    else:
        return False, None, p  # we are at the end
    return v == val, v, p[lv:]

def comp1(p, val):
    if p:
        v = p[0]  # warning: this will produce order dependencies when you spec the parser
    else:
        return False, None, p  # we are at the end
    return v == val, v, p[1:]

def oper(p, func):
    if p:
        v = p[0]
    else:
        return False, None, p  # we are at the end
    return func(v), v, p[1:]

def noneof(string):
    def noneof_(p):
        if not p:
            return True, p, p
        for s in string:
            success, v, rest = comp(p, s, 1)
            if success:
                return False, v, p
        return True, p[0], p[1:]
    return noneof_

def COMP(val):
    lv = len(val)
    if lv == 1:
        def comp1_(p):
            return  comp1(p, val)
        comp1_.__name__ = 'COMP_' + val
        return comp1_
    else:
        def comp_(p):
            return comp(p, val, lv)
        comp_.__name__ = 'COMP_' + val
        return comp_

def EOF(p):
    if p == '':
        return True, '', ''
    else:
        return False, None, p

#
# function to allow implementation of what the parser actually does/outputs

def BOX(v):
    return v,

def RETBOX(v):
    return RETURN((v,))

def RETURNBOX(v):
    if v:
        v, = v
    return RETURN(v)


def FLOP(return_value):
    return RETURN(tuple(return_value[::-1]))

def RETVAL(func, val):
    #def return_bound_val(v):
        #return RETURN(val)
    #return BIND(func, return_bound_val)
    return COMPOSE(func, RETURN(val))

def make_funcs(inpt, lookuptable):
    for token in sorted(set(inpt), key=lambda a: (-len(a), a)):  # sort to simulate right associativity (ie da recognized even if d a token)
        def lookup_function(v):
            return RETURN(lookuptable[v])
        yield BIND(COMP(token), lookup_function)

def jstring(v): return RETURN(''.join(v))

def joinstr(func):
    return BIND(func, jstring)

def STRINGIFY(func):
    return BIND(func, lambda v: RETURN('"' + str(v).replace('"', '\\"') + '"'))

def AT_MOST_ONE(func, fail=True): return BIND(TIMES(func, 0, 1, fail), RETURNBOX)

def EXACTLY_ONE(func, fail=True): return BIND(TIMES(func, 1, 1, fail), RETURNBOX)

# I hate the people who felt the need to make different type blocks for this stuff in 1673
HYPHEN = b'\xe2\x80\x90'.decode()  # Y U DO DIS unicode 2010
EN_DASH = b'\xe2\x80\x93'.decode()  # aka u2013
MINUS_SIGN = b'\xe2\x88\x92'.decode()  # HAH CAUGHT YOU aka u2212
HYPHEN_MINUS = b'\x2d'.decode()  # yes, the thing that most keyboards have
_raw_dash_things = HYPHEN, EN_DASH, MINUS_SIGN, HYPHEN_MINUS
hyphen = COMP(HYPHEN)
minus_sign = COMP(MINUS_SIGN)
en_dash = COMP(EN_DASH)
hyphen_minus = COMP(HYPHEN_MINUS)
_dash_thing = OR(hyphen_minus, en_dash, minus_sign, hyphen)  # THERE ARE TOO MANY AND THEY ALL LOOK THE SAME
dash_thing = RETVAL(_dash_thing, HYPHEN_MINUS)
double_dash_thing = TIMES(dash_thing, 2)
thing_accepted_as_a_dash = BIND(OR(double_dash_thing, dash_thing), lambda v: RETURN(HYPHEN_MINUS))

# basic tokens and operators
thin_space = COMP('\u2009')  # sigh what bloodyminded program would ever produce these!!??!
non_breaking_space = COMP('\xa0')  # nbsp
space = OR(COMP(' '), non_breaking_space, thin_space)
spaces = MANY(space)
spaces1 = MANY1(space)
colon = COMP(':')
exponent = COMP('^')
comma = COMP(',')
point = COMP('.')  # yes this is a period
dot = COMP('·')  # b'\xc2\xb7' MIDDLE DOT U+00B7

newline = COMP('\n')
whitespace_atom = OR(COMP(' '), COMP('\t'), newline)
#whitespace_atom = OR(_whitespace_atom, END(_whitespace_atom, EOF))
whitespace = MANY(whitespace_atom)
whitespace1 = MANY1(whitespace_atom)  # FIXME this is broken to negation? (extremely slow)

CROSS = b'\xc3\x97'.decode()
cross = COMP(CROSS)
x = COMP('x')
X = COMP('X')
by = OR(cross, x, X)  # XXX 2x/day -> twice per day oof

# booleans
_boollookup = {
    'true': True,
    'false': False,
    '#t': True,
    '#f': False,}
bool_word_lower = OR(*make_funcs(_boollookup, _boollookup))
def bool_word_cap(p): return bool_word_lower(p.lower())
boolean = OR(bool_word_lower, bool_word_cap)

# number words
_numlookup = {
    'zero': 0,
    #'half': 0.5,  # better fuzzy
    'one': 1,
    'two': 2,
    'three': 3,
    'four': 4,
    'five': 5,
    'six': 6,
    'seven': 7,
    'eight': 8,
    'nine': 9,
    'ten': 10,
    'twenty': 20,
    'thirty': 30,
    'forty': 40,
    'fifty': 50,
    'sixty': 60,
    'seventy': 70,
    'eighty': 80,
    'ninety': 90,
    # going beyond this requires writing a look-ahead parser
    # for natural language numbers, which is out of scope atm
    'once': 1,
    'twice': 2,
}
num_word_lower = OR(*make_funcs(_numlookup, _numlookup))
def num_word_cap(p): return num_word_lower(p.lower())
num_word = OR(num_word_lower, num_word_cap)

# numbers
digits = [str(_) for _ in range(10)]
def digit(p): return oper(p, lambda d: d in digits)
def char(p): return oper(p, lambda c: c.isalpha())
_int_ = JOINT(TIMES(dash_thing, 0, 1), MANY1(digit), join=True)
int_ = BIND(_int_, lambda i: RETURN(int(''.join(i))))
_float_ = JOINT(TIMES(dash_thing, 0, 1),
                OR(JOINT(MANY1(digit), point, MANY(digit), join=True),
                   JOINT(MANY(digit), point, MANY1(digit), join=True)),
                join=True)
float_ = BIND(_float_, lambda f: RETURN(float(''.join(f))))
E = OR(COMP('E'), COMP('e'))
times = COMP('*')
exponental_notation = JOINT(OR(float_, int_),  # FIXME not including as a num for now because it is sometimes used distributively across plust-or-minus infix
                            COMPOSE(spaces, OR(by, times)),
                            COMPOSE(spaces, COMP('10')),
                            exponent, int_)
_scientific_notation = joinstr(JOINT(joinstr(OR(_float_, _int_)), E, joinstr(_int_)))
scientific_notation = BIND(_scientific_notation, lambda v: RETURN(ScientificNotation(v)))
num = OR(scientific_notation, float_, int_, num_word)  # float first so that int doesn't capture it

def cull_empty(return_value):
    if return_value and not any(return_value[1:]):
            return RETURN(return_value[0])
    return RETURN(return_value)

def flatten(return_value):
    first, rest = return_value
    return RETURN((first, *rest))

def flatten1(return_value):
    return RETURN(tuple(t for r in return_value for t in r))

def noop(function):
    return lambda return_value: RETURN(return_value)



try:
    import rdflib
except:
    log.warning('rdflib not found rdf conversion will not work')


class TypeCaster:
    """ base type converter """

    _bases = {}

    @classmethod
    def cast(cls, thing):
        if isinstance(thing, Converter):
            return thing
        else:
            return cls._bases[type(thing)](thing)

class Converter:

    def __mul__(self, other):
        if not isinstance(self, other.__class__):
            return other.__mul__(self)
        else:
            return super().__mul__

    def asRdf(self):
        return rdflib.Literal(self)


class boolc(Converter, float):
    pass


class floatc(Converter, float):
    def __truediv__(self, other):
        return floatc(super().__truediv__(other))


class intc(Converter, int):
    def __truediv__(self, other):
        return floatc(super().__truediv__(other))


class strc(Converter, str):
    def __add__(self, other):
        return self.__class__(super().__add__(other))


TypeCaster._bases[float] = floatc
TypeCaster._bases[int] = intc
TypeCaster._bases[str] = strc



# types that can be imported quickly


class Unit:
    """ sigh python """


class Quantity:
    """ sigh python """
    tag = 'quantity'


class Measurement:
    """ sigh python """
    tag = 'measurement'


class Range:
    """ sigh python """
    tag = 'range'


class ProtcurExpression:
    """ sigh python """


class AJ:
    """ As json helper """

    @property
    def rchildren(self):  # FIXME this does not go here
        if hasattr(self, 'body'):
            for child in self.body:
                yield child
                yield from child.rchildren

    def asJson(self):
        # TODO
        if self.prov is None:
            breakpoint()
        else:
            if hasattr(self.prov, 'id'):
                # hypothesis helper case
                # reasonable compromise with using curie
                # FIXME evil because it conflates the
                # hypothesis view with the protcur view
                # could use them as blanknodes but tricky then
                # FIXME this should not be on the generic AJ class ...
                id = 'hyp-protcur:' + self.prov.id
            elif isinstance(self.prov, str):
                # string identifier case
                id = self.prov
            #elif hasattr(self, '_id'):
                #id = self._id
            else:
                # FIXME
                # is ast node ...
                # which is kind of AJ?
                # also consider using _id?
                id = self._value
                breakpoint()

        #context = {  # doesn't go here
            #'children': {'@id': 'ilxtr:protcurBody' , '@type': '@id'},
        #}

        out = {
            #'@context': context,
            '@id': id,
            '@type': self._type,
            'node_type': self.__class__.__name__,  # TODO @type is what?
        }

        value = self._value
        if isinstance(value, str) and value.startswith('aspect-raw:'):  # FIXME hardcoding
            out['value'] = value
        elif (isinstance(value, str) or
              isinstance(value, int) or
              isinstance(value, float)):
        #elif not (isinstance(value, dict) or
                  #isinstance(value, list) or
                  #value.__class__.__name__ == 'Term'  # FIXME DUMB
                  #):
            out['raw_value'] = {'@value': value}
        else:
            out['ast_value'] = value

        if 'raw_value' not in out and hasattr(value, 'original'):
            # lift original values back to their original position in the ast
            out['raw_value'] = {'@value': value.original}

        if self.prov and hasattr(self.prov, 'id'):  # FIXME sigh assumptions about prov
            uris = {
                'uri_human_context': self.prov.shareLink,
                'uri_human_static': self.prov.htmlLink,
                'uri_api': 'https://hypothes.is/api/annotations/' + self.prov.id,
            }
            out.update(uris)
            if self.__class__.__name__ in ('Parameter', 'Invariant'):  # FIXME sigh
                out['raw_value'] = {'@value': self.prov._value()}
                if (self.rest and
                    (isinstance(self.rest[0], list) or isinstance(self.rest[0], tuple)) and
                    self.rest[0] and
                    self.rest[0][0] == "rest"):
                    # TODO mark out as a whole as an error probably
                    out['rest'] = self.rest[0][1]

        if hasattr(self, 'body'):
            if self.body:
                try:
                    body = [node.asJson() for node in self.body]  # FIXME order semantics
                except AttributeError as e:
                    breakpoint()
                    'asdf'
                out['children'] = [b['@id'] for b in body if '@id' in b]

        # TODO nest children in the jsonld and then hope it flattens?
        return out

    def __str__(self):
        """ you proably shouldn't use this ... """
        return str(self._value)



import pytest
import unittest
from .common import *
from pysercomb.parsers import units
from pysercomb.pyr import units as pyru


class TestForms(unittest.TestCase):

    def test_weird(self):  # TODO
        twice = parameter_expression('2x/day')

        eq = parameter_expression('mEq/mL')  # was just missing equivalents

        w0 = parameter_expression('10-to-12')
        w1 = parameter_expression('10 -to- 12')
        w2 = parameter_expression('10 - 12')
        w3 = parameter_expression('10 mg·kg−1·h−1')  # dot for multiplication
        w4 = parameter_expression('0.4V ~ 1.5V')  # infix approx

        w5 = parameter_expression('10 per day')
        w6 = parameter_expression('10 / day')
        w7 = parameter_expression('10per day')
        w8 = parameter_expression('10/ day')

        n0 = parameter_expression('1,2,3')  # not valid for commas
        n1 = parameter_expression('1000,000')  # not valid for commas
        n2 = parameter_expression('100,00')
        assert parameter_expression('1,000') == (True, ('param:quantity', 1000, ()), '')
        assert parameter_expression('100,000') == (True, ('param:quantity', 100000, ()), '')

    def test_ratio(self):
        assert parameter_expression('1:1000')[1] == ('param:ratio', 1, 1000), 'ratio failed'
        assert parameter_expression('2:1')[1] == ('param:ratio', 2, 1), 'ratio failed'

    def test_time_seconds(self):
        assert (parameter_expression('00:00:01')[1]
                == ('param:quantity', 1, ('param:unit', ('quote', 'seconds')))), 'duration failed'

    def test_time_minutes(self):
        assert (parameter_expression('00:01:01')[1]
                == ('param:quantity', 61, ('param:unit', ('quote', 'seconds')))), 'duration failed'

    def test_time_hours(self):
        assert (parameter_expression('01:01:01')[1]
                == ('param:quantity', 3661, ('param:unit', ('quote', 'seconds')))), 'duration failed'

    def test_time_weird(self):
        assert (parameter_expression('1:99:01')[1]
                == ('param:quantity', 9541, ('param:unit', ('quote', 'seconds')))), 'duration failed'

    def test_iso8601duration(self):
        # when you stop to think about it, it is kind of silly to report a
        # duration that crosses the months/days boundary due to the fact that
        # months have a variable number of days, so you will already be +/- a
        # day or 3 depending on the exact place in time where the duration is
        # specified, for use cases outside scientific I can image that it might
        # be useful
        #('P10D12H'),  # TODO should fail
        strings = (
            ('P1.5M', ('iso8601-duration-date',
                       ('param:quantity', 1.5, ('param:unit', ('quote', 'months'))),)),
            ('P1,5M', ('iso8601-duration-date',
                       ('param:quantity', 1.5, ('param:unit', ('quote', 'months'))),)),
            ('P1M', ('iso8601-duration-date',
                     ('param:quantity', 1, ('param:unit', ('quote', 'months'))),)),
            ('PT1M', ('iso8601-duration-time',
                      ('param:quantity', 1, ('param:unit', ('quote', 'minutes'))),)),
            ('P10D', ('iso8601-duration-date',
                      ('param:quantity', 10, ('param:unit', ('quote', 'days'))),)),
            ('PT12H', ('iso8601-duration-time',
                       ('param:quantity', 12, ('param:unit', ('quote', 'hours'))),)),
            ('P1Y1D', ('iso8601-duration-date',
                       ('param:quantity', 1, ('param:unit', ('quote', 'years'))),
                       ('param:quantity', 1, ('param:unit', ('quote', 'days'))),)),
            ('PT1M1S', ('iso8601-duration-time',
                        ('param:quantity', 1, ('param:unit', ('quote', 'minutes'))),
                        ('param:quantity', 1, ('param:unit', ('quote', 'seconds'))),)),
            ('P10DT12H', ('iso8601-duration-datetime',
                          ('param:quantity', 10, ('param:unit', ('quote', 'days'))),
                          ('param:quantity', 12, ('param:unit', ('quote', 'hours'))),)),
            ('P4Y3M10DT12H1M2S', ('iso8601-duration-datetime',
                                  ('param:quantity', 4, ('param:unit', ('quote', 'years'))),
                                  ('param:quantity', 3, ('param:unit', ('quote', 'months'))),
                                  ('param:quantity', 10, ('param:unit', ('quote', 'days'))),
                                  ('param:quantity', 12, ('param:unit', ('quote', 'hours'))),
                                  ('param:quantity', 1, ('param:unit', ('quote', 'minutes'))),
                                  ('param:quantity', 2, ('param:unit', ('quote', 'seconds'))),)),
        )
        bads = []
        for s, r in strings:
            ok, res, rest = iso8601duration(s)
            if r != res:
                print('exp:', r)
                print('got:', res)
                bads.append(s)

        assert not bads, bads

    def test_postnatal_day_vs_iso8601duration(self):
        assert (parameter_expression('P14')[1]
                # FIXME -day vs -days
                == ('param:quantity', 14,
                    ('param:prefix-unit', ('quote', 'postnatal-days')))), 'postnatal_day failed'
        assert (parameter_expression('P14D')[1]
                == ('iso8601-duration-date',
                    ('param:quantity', 14,
                     ('param:unit', ('quote', 'days'))))), 'iso8601duration failed'


class TestUnit(unittest.TestCase):

    def test_micron(self):
        assert unit('micron') == unit('um'), 'micron fail'

    def test_mmHg(self):
        expect = (True, ('param:unit', ('quote', 'millimeters-hg')), '')
        assert unit('mmHg') == expect, 'mmHgf'

    def test_kg_Kg(self):
        expect = (True, ('param:unit', ('quote', 'grams'), ('quote', 'kilo')), '')
        assert unit('kg') == expect, 'kgf'
        # if you can't get your SI prefixes correct I can't help you
        assert unit('Kg') == (True, ('param:unit', ('quote', 'kelvin')), 'g'), 'Kgf'

    def test_siprefix(self):
        assert debug_dict['siprefix']('milli') == (True, ('quote', 'milli'), ''), 'oops'
        assert debug_dict['siprefix']('deci') == (True, ('quote', 'deci'), ''), 'oops'

    def test_mm(self):
        expect = (True, ('param:unit', ('quote', 'meters'), ('quote', 'milli')), '')
        assert unit('mm') == expect, 'mmf'
        assert unit('mmeters') == expect, 'mmf'
        assert unit('millimeters') == expect, 'mmf'

    def test_becq(self):
        """ plural issues """
        expect = (True, ('param:unit', ('quote', 'becquerels')), '')
        assert unit('becquerels') == expect, 'becqf'
        assert unit('becquerel') == expect, 'becqf'
        assert unit('Bq') == expect, 'becqf'

    def test_minutes(self):
        """ make sure we don't parse min -> milli inches """
        expect = (True, ('param:unit', ('quote', 'minutes')), '')
        assert unit('minutes') == expect, 'minutes did not parse to minutes'
        assert unit('minute') == expect, 'minute did not parse to minutes'
        assert unit('min') == expect, 'min did not parse to minutes'

    def test_inches(self):
        assert unit('in') == (True, ('param:unit', ('quote', 'inches')), ''), 'in did not parse to inches'

    def test_ohms(self):
        assert unit('R') == (True, ('param:unit', ('quote', 'ohms')), ''), 'R did not parse to ohms'

    def test_rcf(self):
        assert unit('RCF') == (True, ('param:unit', ('quote', 'relative-centrifugal-force')), ''), 'RCF did not parse to relative-centrifugal-force'

    def test_newtons(self):
        assert unit('N') == (True, ('param:unit', ('quote', 'newtons')), ''), 'N did not parse to newtons'

    def test_numerical_aperture(self):
        msg = 'NA did not parse to numerical-aperture'
        assert unit('NA') == (True, ('param:unit', ('quote', 'numerical-aperture')), ''), msg

    def test_fold(self):
        msg = 'Ax failed to parse as fold!'
        assert parameter_expression('40x')[1] == ('param:quantity', 40, ('param:unit', ('quote', 'fold'))), msg

    def test_dimensions(self):
        msg = 'A x B failed to parse as dimensions!'
        # 50 x 50 um is ambiguous, unless additional information can be supplied
        # if you have the additional context then use dimensions_no_math
        assert parameter_expression('50 um x 50 um')[1] == ('param:dimensions',
                                                            ('param:quantity', 50,
                                                             ('param:unit', ('quote', 'meters'), ('quote', 'micro'))),
                                                            ('param:quantity', 50,
                                                             ('param:unit', ('quote', 'meters'), ('quote', 'micro')))), msg

    def test_dimensions_no_math(self):
        assert dimensions_no_math('50 x 50 um')[1] == ('param:dimensions',
                                                       ('param:quantity', 50, ()),
                                                       ('param:quantity', 50,
                                                        ('param:unit', ('quote', 'meters'), ('quote', 'micro')))), msg



    def test_percent(self):
        msg = '% failed to parse'
        assert parameter_expression('0.3%')[1] == ('param:quantity', 0.3, ('param:unit', ('quote', 'percent'))), msg

    def test_mm_range(self):
        expect = (True, ('param:quantity', ('param:expr', ('range', 5, 6)), ('param:unit', ('quote', 'meters'), ('quote', 'milli'))), '')
        assert parameter_expression('5-6mm') == expect, 'oops'
        assert parameter_expression('5-6 mm') == expect, 'oops'
        assert parameter_expression('5-6 millimeters') == expect, 'oops'

    def test_exp_vs_range(self):
        # range is fighting with exp_short
        # There is no good solution at this stage of the pipeline.
        # The issue is that a range separator without spaces can be parsed as a negative exponent.
        # Some authors like to put 1%-2% doubling the percent unit.
        # Percent does not combine with exp_short, which we can fix at the cost of some complexity,
        # but the underlying issue remains, e.g. with the 1kg-2kg example below.
        # One possible solution is to ensure that exp_short is not followed by another unit, which
        # is not viable because kg-1m is technically a valid unit, if horribly mangled I think.
        # Thus the best way forward is to check to see if there is a rest in the calling context
        # and if there is and there is a dash_thing in the input then should reparse with the
        # dash_thing surrounded by spaces, in which case the range will parse correctly and
        # there should be no rest value.
        o0 = unit('%-1')
        parsing._dobreak = True
        o1 = parameter_expression('1%-1')
        o5 = parameter_expression('1%-1%')

        k0 = parameter_expression('1kg-2kg')
        k1 = parameter_expression('1kg - 2kg')

        m0 = parameter_expression('1mm3-2mm3')
        m1 = parameter_expression('1mm3 - 2mm3')

        o2 = parameter_expression('1%+1')
        o3 = parameter_expression('1%*1')
        o4 = parameter_expression('1%/1')
        parsing._dobreak = False

    def test_molar_space(self):
        out = pyru.UnitsParser(' 0.1M')
        assert out == ('param:quantity', 0.1, ('param:unit', ('quote', 'molarity')))

    def test_unit_exp(self):
        res = exp_short('mm^3')
        _, out, _ = res
        assert out == ('^', ('param:unit', ('quote', 'meters'), ('quote', 'milli')), 3), res

        res = unit_thing('mm^3')
        _, out, _ = res
        assert out == ('^', ('param:unit', ('quote', 'meters'), ('quote', 'milli')), 3), res

        # should fail
        #res = unit_expr_atom('mm^3')
        #_, out, _ = res
        #assert out == ('^', ('param:unit', ('quote', 'meters'), ('quote', 'milli')), 3), res

        #res = unit('mm^3')

        res = unit_expression('mm^3')
        _, out, _ = res
        assert out == ('param:unit-expr', ('^', ('param:unit', ('quote', 'meters'), ('quote', 'milli')), 3)), res

        res = suffix_unit('mm^3')
        _, out, _ = res
        assert out == ('param:unit-expr', ('^', ('param:unit', ('quote', 'meters'), ('quote', 'milli')), 3)), res

    def test_implicit_count_ratio(self):
        ok, out, rest = suffix_unit(' / mm^3')
        assert out == ('param:unit-expr',
                       ('/', ('param:unit', ('quote', 'count')),
                        ('^', ('param:unit', ('quote', 'meters'), ('quote', 'milli')), 3))), (ok, out, rest)

    def test_parens(self):
        tests = (('s * m', ()),
                 ('s * mm', ()),
                 ('s * m^2', ()),
                 ('s * m ^ 2', ()),

                 ('s * m', ()),
                 ('s / m', ()),

                 ('mg kg–1', ()),
                 ('mg * kg–1', ()),
                 ('(J * m ) / s', ()),
                 ('(m * g) / s', ()),
                 ('s / m * g * l', ()),
                 ('M / (A * V * R)', ()),

                 ('as * m * g * l', ()),
                 ('fs / m * g / l', ()),

                 ('ps * m * (g * l)', ()),
                 ('ns / m * (g / l)', ()),

                 ('(s * m) * g * l', ()),
                 ('(s / m) * g / l', ()),

                 ('(s * m) * (g * l)', ()),
                 ('(s / m) * (g / l)', ()),
                 ('((s / m) * (g / l))', ()),

                 ('(m * g) / s^2', ()),
                 ('( m * g ) / s^2', ()),
                 #'(( m * g ) / s^2)',
                 ('(m * g) / s', ()),

                 ('(((m * g) * s) * J)', ()),
                 ('(m * (g * (s * J)))', ()),
        )
        ueatest = [unit_expr_atom(t) for t, e in tests]
        ueabad = [r for r in ueatest if r[-1]]
        uetest = [unit_expr(f'({t})') for t, e in tests]
        uebad = [r for r in uetest if r[-1]]
        ue1test = [unit_expr(f'{t}') for t, e in tests]
        ue1bad = [r for r in ue1test if r[-1]]
        uentest = [unit_expression(t) for t, e in tests]
        uenbad = [r for r in uentest if r[-1]]
        assert not uenbad


class TestNumExpr(unittest.TestCase):

    func = staticmethod(num_expression)

    def test_0(self):
        res = self.func('0')
        _, out, _ = res
        assert out == 0, res

    def test_1(self):
        res = self.func('0 + 0')
        _, out, _ = res
        assert out == ('param:expr', ('+', 0, 0)), res

    def test_2(self):
        res = self.func('(2 + 1)')
        _, out, _ = res
        assert out == ('param:expr', ('+', 1, 2)), res

    def test_3(self):
        res = self.func('0 + 0 + 0')
        _, out, _ = res
        assert out == ('param:expr', ('+', 0, 0, 0)), res

    def test_3_0(self):
        res = self.func('(0 + 0) + 0')
        _, out, _ = res
        assert out == ('param:expr', ('+', 0, 0, 0)), res

    def test_3_1(self):
        res = self.func('0 + (0 + 0)')
        _, out, _ = res
        assert out == ('param:expr', ('+', 0, 0, 0)), res

    def test_4(self):
        res = self.func('(0 + 0 + 0)')
        _, out, _ = res
        assert out == ('param:expr', ('+', 0, 0, 0)), res

    def test_5(self):
        """ breaks the parser """
        res = self.func('((0 + 0) + 0)')
        _, out, _ = res
        assert out == ('param:expr', ('+', 0, 0, 0)), res

    def test_6(self):
        res = self.func('((0 + 0) + 0)')
        _, out, _ = res
        assert out == ('param:expr', ('+', 0, 0, 0)), res

    def test_7(self):
        res = self.func('((0 + 0) + 0)')
        _, out, _ = res
        assert out == ('param:expr', ('+', 0, 0, 0)), res

    def test_8(self):
        res = self.func('(25)')
        _, out, _ = res
        assert out == 25, res

    def test_8_0(self):
        res = self.func('-25')
        _, out, _ = res
        assert out == -25, res

    def test_9(self):
        res = self.func('(-25)')
        _, out, _ = res
        assert out == -25, res

    def test_num_expression(self):
        tests = (
            ('1', (True, 1, '')),
            ('1 + 2', (True, ('param:expr', ('+', 1, 2)), '')),
            ('(1 + 2)', (True, ('param:expr', ('+', 1, 2)), '')),
        )
        ntest = [(t, num_expression(t), e) for t, e in tests]
        nbad = [(t, r, e) for t, r, e in ntest if r != e]
        assert not nbad, '\n' + '\n'.join([f'{t}' for t in nbad])

    def test_parens(self):
        test = '(10 + 3) * 4'  # FIXME should be able to subsum all of this into a single expr
        _, out, _ = num_expression(test)
        assert out == ('param:expr', ('*', 4, ('+', 3, 10)))

    def test_infix_expr_many(self):
        text = '1 * 3 * 2 * 4'
        out = num_expression(text)
        test = (True,
                ('param:expr',
                 ('*',
                  1,
                  2,
                  3,
                  4),),
                '')
        assert out == test

    def test_infix_expr_op_order_plus_mult(self):
        text = '1 * 2 + 4 * 3'
        out = num_expression(text)
        test = (True,
                ('param:expr',
                 ('+',
                  ('*',
                   1,
                   2),
                  ('*',
                   3,
                   4)),),
                '')
        assert out == test

    def test_infix_expr_op_order_plus_mult_2(self):
        text = '2 * 1 + 4 * 5 * 3'
        out = num_expression(text)
        test = (True,
                ('param:expr',
                 ('+',
                  ('*',
                   1,
                   2),
                  ('*',
                   3,
                   4,
                   5)),),
                '')
        assert out == test

    def test_infix_expr_noncommutative(self):
        """ do not reorder across non-commutative functions """
        text = '3 ^ 2'  # can't use subtraction since we don't parse it
        out = num_expression(text)
        test = (True,
                ('param:expr', ('^', 3, 2),),
                '')
        assert out == test


class TestExpr(unittest.TestCase):
    def test_mixed_expr(self):
        tests = (
            ('(1 + 2) * (3 + 4)', ()),
            ('1 / mm^3', ()),
            ('(1 + 2) mm', ()),
            ('(3 + 4) mm^3', ()),
            ('(0 + 0) / mm^3', ()),
            ('(0 + (0 + 0))', ()),
            ('1 / 2 + 3 - 4', ()),
            ('1 / (2 + 3) - 4', ()),
            ('(1 / (2 + 3) - 4)', ()),
            ('(1 / (2 + 3) - 4) mm', ()),
            ('(1 / (2 + 3) - 4) mm^3', ()),
            ('(1 / (2 + 3) - 4) / mm^3', ()),
        )
        #qtest = [quantity(t) for t, e in tests]
        #qbad = [r for r in qtest if r[-1]]
        #assert not qbad, qbad
        #entest = [expression(t) for t, e in tests]
        #enbad = [r for r in entest if r[-1]]

        pentest = [parameter_expression(t) for t, e in tests]
        penbad = [r for r in pentest if r[-1]]
        assert not penbad, '\n' + '\n'.join([f'{t}' for t in penbad])

    def test_mixed_unit_op_order_simple(self):
        test = '1 / mm^3'
        _, out, _ = parameter_expression(test)
        print(out)
        assert out == ('param:quantity',
                       1,
                       ('param:unit-expr',
                        ('/', ('param:unit', ('quote', 'count')),
                         ('^', ('param:unit', ('quote', 'meters'), ('quote', 'milli')), 3))))

    @pytest.mark.skip('not fixed yet')
    def test_mixed_seconds(self):
        test = '1 / s'
        _, out, _ = parameter_expression(test)
        print(out)
        assert out == ('param:quantity',
                       1,
                       ('param:unit-expr',
                        ('/', ('quote', 'seconds'))))

    def test_mixed_unit_op_order(self):
        test = '4.7 +- 0.6 x 10^7 / mm^3'
        test = '(4.7 +- 0.6 x 10^7) / mm^3'
        _, out, _ = parameter_expression(test)
        print(out)
        assert out == ('param:quantity',
                       ('param:expr',
                        ('*',  # FIXME mark as param:expr?
                         ('plus-or-minus', 4.7, 0.6),
                         ('^', 10, 7))),
                       ('param:unit-expr',
                        ('/', ('param:unit', ('quote', 'count')),
                         ('^', ('param:unit', ('quote', 'meters'), ('quote', 'milli')), 3))))

    def test_prefix_infix_expr(self):
        text = '~1 - 3 mm'
        out = prefix_expression(text)
        test = (True,
                (('approximately',
                  ('param:quantity',
                   ('param:expr', ('range', 1, 3)),
                   ('param:unit', ('quote', 'meters'), ('quote', 'milli')))
                   ),),
                '')
        assert out == test

    def test_prefix_expr(self):
        text = '>1'
        out = prefix_expression(text)
        test = (True,
                (('>', ('param:quantity', 1, ())),),
                '')
        assert out == test

    def test_float_scinote(self):
        test = '0.000002 molarity'
        out = parameter_expression(test)
        assert out == (True,
                       ('param:quantity',
                        2e-06,
                        ('param:unit', ('quote', 'molarity'))),
                       '')


class TestMain(unittest.TestCase):
    def test_main(self):
        """ catch any changes here as well, eventually we should remove main for this ... """
        units.main()



from protcur.config import __units_folder__ as units_path
from pysercomb.parsers.units import get_unit_dicts, make_unit_parser
from pysercomb.types import intc, strc
from . import types as intf


def chain(*tups):
    for t in tups:
        yield from reversed(t)  # reversed because the piority ordering is inverted


class ImplFactoryHelper:
    # FIXME NOTE
    # UnitsParser is a case where
    @classmethod
    def bindImpl(cls, class_name, *classes, **kwargs):
        """ returns a new interpreter bound to a particular implementation of
            Unit

            specify the underlying python classes that the parser targets
            this is sort of like being able to change the #lang you are using """
        if class_name is None:
            for cls_ in classes:
                setattr(cls, '_' + cls_.__name__, cls_)
            for name, some_callable in kwargs.items():
                setattr(cls, '_' + name, some_callable)

        else:
            class_dict = {'_' + cls.__name__:cls for cls in classes}
            return type(class_name, (cls,), class_dict)


class UnitsHelper:

    __setup = False

    @staticmethod
    def setup():
        """ call setup on UnitsHelper once, not on subclasses """

        if hasattr(UnitsHelper, '__setup') and UnitsHelper.__setup:
            return  # already setup

        UnitsHelper.__setup = True

        dicts = get_unit_dicts(units_path)

        (parameter_expression, quantity, unit, *_,
         debug_dict) = make_unit_parser(dicts=dicts)

        UnitsHelper._parameter_expression = staticmethod(parameter_expression)

        gs = globals()
        for dict_ in dicts:
            gs.update(dict_)

        UnitsHelper.si_exponents = {prefix if prefix is None else strc(prefix):intc(exp)
                            for prefix, exp in prefixes_si_exponents}
        UnitsHelper.si_exponents_inv = {e:p for p, e in UnitsHelper.si_exponents.items()}

        UnitsHelper.unit_dict = {strc(unit):strc(abbrev) for abbrev, unit in
                          chain(units_si,
                                units_extra,
                                units_extra_prefix,
                                units_dimensionless,
                                units_dimensionless_prefix,
                                units_imp,)}

        UnitsHelper.unit_dict_inv = {a:u for u, a in UnitsHelper.unit_dict.items()}

        # don't actually need this because its in the ast
        #prefix_units = set(unit for abbrev, unit in
                        #chain(units_extra_prefix,
                                #units_dimensionless_prefix))

        UnitsHelper.prefix_dict = {strc(prefix):strc(abbrev) for abbrev, prefix in prefixes_si}
        UnitsHelper.prefix_dict_inv = {p:a for a, p in UnitsHelper.prefix_dict.items()}

        # surely there is a more elegant way ...
        # TODO
        UnitsHelper.conversion = {'__truediv__':[],
                                  '__mul__':[]}


class Expr(intf.ProtcurExpression, UnitsHelper, ImplFactoryHelper):
    op = None

    def __add__(self, other):
        return self._Add(self, other)

    def __iadd__(self, other):
        return self.__add__(other)

    #def __mul__(self, other):
        #return self._Mul(self, other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._Div(self, other)

    def __pow__(self, other):
        return self._Exp(self, other)

    def __lt__(self, other):
        if other is LessThan:
            return LessThan(None, other)
        else:
            raise NotImplementedError

    def __gt__(self, other):
        if other is GreaterThan:
            return GreaterThan(None, other)
        else:
            raise NotImplementedError

    #@express
    #def asRdf(self):
        #yield from self.asRdf(rdflib.BNode())

    @property
    def ttl(self):
        graph = rdflib.Graph()
        OntCuries.populate(graph)
        [graph.add(t) for t in self.asRdf()]
        return graph.serialize(format='nifttl')


# comparison classes

class _Than(intf.ProtcurExpression):
    op = None
    tag = None

    def json(self):
        # FIXME prefix vs suffix quantities
        return dict(type=self.tag, value=self.right, unit=self.right.units.json())

    @classmethod
    def fromJson(cls, json):
        assert json['type'] == cls.tag
        #return cls(None, json['value'], json['units'])
        return cls(None, json['value'])  # FIXME units not supported yet

    def to_base_units(self):
        return self.__class__(
            *(_ if _ is None else _.to_base_units() for _ in (self.left, self.right)))

    @property
    def dimensionality(self):
        ld, rd = None, None
        if self.left is not None:
            ld = self.left.dimensionality

        if self.right is not None:
            rd = self.right.dimensionality

        if ld and rd and ld != rd:  # FIXME should catch during construction ya?
            msg = f'{self.__class__.__name__} dimensionality mismatch! {ld} != {rd}'
            raise TypeError(msg)

        return rd if ld is None else ld

    @property
    def units(self):
        lu, ru = None, None
        if self.left is not None:
            lu = self.left.units

        if self.right is not None:
            ru = self.right.units

        if lu and ru and lu != ru:  # FIXME should catch during construction ya?
            msg = f'{self.__class__.__name__} units mismatch! {lu} != {ru}'
            raise TypeError(msg)

        return ru

    def __hash__(self):
        return hash((self.__class__, self.left, self.right))

    def __eq__(self, other):
        # FIXME WARNING assuming that right is always what we wind up with
        return type(self) == type(other) and self.right == other.right

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other

    def __init__(self, left, right):
        self.left = left
        self.right = right


    def __call__(self, other):
        if self.left is None:
            return getattr(other, self._op)(self)

    def __str__(self):
        left = f'{self.left} ' if self.left else ''
        return f'{left}{self.op} {self.right}'


class ltclass(type):
    _op = '__lt__'
    def __lt__(self, other):
        return self(None, other)


class gtclass(type):
    _op = '__gt__'
    def __gt__(self, other):
        return self(None, other)


class LessThan(_Than, metaclass=ltclass):
    op = '<'
    tag = 'less-than'

    def __lt__(self, other):
        if type(self) == type(other):
            return self.right < other.right
        else:
            # NOTE a bit strange, but
            # < 18 hours can be < 20 minutes
            # of course one might say that on
            # average it is most certainly
            # greater than 20 minutes, but
            # we weren't given a lower bound
            return True

    def __gt__(self, other):
        if type(self) == type(other):
            return self.right > other.right
        else:
            # less thans can never be greater than something else
            # because in theory their value can always be less than
            # any obtained value ... now, in reality most of the time
            # that people use < 18 hours, they are implying that there
            # is a lower bound of zero hours
            return False


class GreaterThan(_Than, metaclass=gtclass):
    op = '>'
    tag = 'greater-than'

    def __lt__(self, other):
        if type(self) == type(other):
            return self.right < other.right
        else:
            return False

    def __gt__(self, other):
        if type(self) == type(other):
            return self.right > other.right
        else:
            return True


class LessThanOrEqual(LessThan):
    op = '<='
    tag = 'less-than-or-equal'
    # TODO check the sorting rules for this


class GreaterThanOrEqual(GreaterThan):
    op = '>='
    tag = 'greater-than-or-equal'
